/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "winbase.h"
#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define GL_SetupViewport(x)

static inline void
GL_draw_vertices(GL_D3D_priv* priv, GLenum gltype, DWORD fvf,
		 DWORD start, DWORD count,
		 GL_IndexFunc index, LPVOID index_data)
{
    if (priv->d3dp->renderstate[D3DRENDERSTATE_LIGHTING] && (fvf & D3DFVF_NORMAL || priv->d3d_version >= 8))
	GL_draw_unlit_vertices(priv, gltype, fvf, start, count, index, index_data);
    else assert(FALSE);
}

static DWORD GL_DirectIndex(DWORD i, LPVOID p)
{
    return i;
}

static DWORD GL_IndexedIndex(DWORD i, LPVOID p)
{
    return ((LPWORD)p)[i];
}

void GL_GetPrimitiveType(LPD3DHAL_DP2COMMAND cmd,
				GLenum* lppt, LPDWORD lpdwCount,
				LPBOOL lpbIndexed)
{
    DWORD dwCount;
    GLenum pt;
    BOOL indexed = FALSE;

    switch (cmd->bCommand)
    {
    case D3DDP2OP_POINTS:
	/* Special case. */
	TRACE("Draw POINTS\n");
	pt = GL_POINTS;
	dwCount = 0;
	indexed = FALSE;
	break;

    case D3DDP2OP_INDEXEDLINELIST2:
	indexed = TRUE;
    case D3DDP2OP_LINELIST:
	TRACE("Draw LINES\n");
	pt = GL_LINES;
	dwCount = cmd->u.wPrimitiveCount*2;
	break;

    case D3DDP2OP_INDEXEDLINESTRIP:
	indexed = TRUE;
    case D3DDP2OP_LINESTRIP:
	TRACE("Draw LINE_STRIP\n");
	pt = GL_LINE_STRIP;
	dwCount = cmd->u.wPrimitiveCount+1;
	break;

    case D3DDP2OP_INDEXEDTRIANGLELIST2:
	indexed = TRUE;
    case D3DDP2OP_TRIANGLELIST:
	TRACE("Draw TRIANGLES\n");
	pt = GL_TRIANGLES;
	dwCount = cmd->u.wPrimitiveCount*3;
	break;

    case D3DDP2OP_INDEXEDTRIANGLESTRIP:
	indexed = TRUE;
    case D3DDP2OP_TRIANGLESTRIP:
	TRACE("Draw TRIANGLE_STRIP\n");
	pt = GL_TRIANGLE_STRIP;
	dwCount = cmd->u.wPrimitiveCount+2;
	break;

    case D3DDP2OP_INDEXEDTRIANGLEFAN:
	indexed = TRUE;
    case D3DDP2OP_TRIANGLEFAN:
	TRACE("Draw TRIANGLE_FAN\n");
	pt = GL_TRIANGLE_FAN;
	dwCount = cmd->u.wPrimitiveCount+2;
	break;

    default:
	ERR("Unhandled primitive\n");
	pt = 0;
	dwCount = 0;
	indexed = FALSE;
	break;
    }

    *lpdwCount = dwCount;
    *lppt = pt;
    *lpbIndexed = indexed;
}

HRESULT D3D_GL_draw(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    const WORD* indices;
    DWORD fvf;
    DWORD dwCount;
    DWORD dwIndexBase;
    DWORD dwVStart, dwVCount;
    GLenum pt;
    BOOL indexed;

    GL_GetPrimitiveType(cmd, &pt, &dwCount, &indexed);

    indices = indexed ? data2 : NULL;
    dwIndexBase = indexed ? ((LPD3DHAL_DP2STARTVERTEX)data1)->wVStart : 0;
    dwVStart = priv->dp2_dwVertexOffset;
    dwVCount = priv->dp2_dwVertexLength;

    /* Assume we can until proven otherwise. */
    priv->glFlags |= GLF_USE_VERTEX_ARRAYS;

    if (priv->glFlags & GLF_VERTEX_PROGRAM) {
	fvf = 0;
	/* load up vertex program stuff */
	D3D_GL_setup_programmable(priv, data1, indices, pt, dwCount, indexed);
	
    } else {
	fvf = D3D_GL_setup_fixed(priv, data1, indices, pt, dwCount, indexed);
	if (!fvf) return E_FAIL;
    }

    TRACE("start\n");

    if (pt == GL_POINTS)
    {
	LPD3DHAL_DP2POINTS dat = (LPD3DHAL_DP2POINTS)data2;
	DWORD p;

	for (p=0; p<cmd->u.wPrimitiveCount; p++,dat++) {
	    if (priv->glFlags & GLF_USE_VERTEX_ARRAYS)
		glDrawArrays(pt, dat->wVStart, dat->wCount);
	    else
		GL_draw_vertices(priv, pt, fvf, dat->wVStart, dat->wCount,
				 GL_DirectIndex, NULL);
	}
    }
    else if (indexed)
    {
	if (priv->glFlags & GLF_USE_VERTEX_ARRAYS) {
#if 1
	    glDrawRangeElements(pt, dwVStart - dwIndexBase,
				dwVStart - dwIndexBase + dwVCount,
				dwCount, GL_UNSIGNED_SHORT, indices);
#else
	    glDrawElements(pt, dwCount, GL_UNSIGNED_SHORT, indices);
#endif
	} else
	    GL_draw_vertices(priv, pt, fvf, 0, dwCount,
			     GL_IndexedIndex, data2);
    }
    else
    {
	LPD3DHAL_DP2STARTVERTEX dat = (LPD3DHAL_DP2STARTVERTEX)data1;

	if (priv->glFlags & GLF_USE_VERTEX_ARRAYS)
	    glDrawArrays(pt, dwVStart + dat->wVStart, dwCount);
	else
	    GL_draw_vertices(priv, pt, fvf, dwVStart + dat->wVStart, dwCount,
			     GL_DirectIndex, NULL);
    }

    CHECK_GL_ERROR("DrawPrimitive");

    if (priv->glFlags & GLF_VERTEX_PROGRAM) {
        /* finish up vertex program stuff */
        D3D_GL_finish_programmable(priv);
    }
    else {
	D3D_GL_finish_fixed(priv);
    }

    TRACE("done\n");
    return D3D_OK;
}

HRESULT D3D_GL_clear(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2CLEAR dat = (LPD3DHAL_DP2CLEAR)data1;
    GLboolean ztest = GL_FALSE;
    GLbitfield mask = 0;

    /* For the moment, ignore the rectangles */

    TRACE("Flags=%8lx\n", dat->dwFlags);

    GL_SetupViewport(This);

    if (dat->dwFlags & D3DCLEAR_TARGET) {
	mask |= GL_COLOR_BUFFER_BIT;
	glClearColor(RGBA_GETRED(dat->dwFillColor) / 255.0,
		     RGBA_GETGREEN(dat->dwFillColor) / 255.0,
		     RGBA_GETBLUE(dat->dwFillColor) / 255.0,
		     RGBA_GETALPHA(dat->dwFillColor) / 255.0);
    }
    if (dat->dwFlags & D3DCLEAR_STENCIL) {
	mask |= GL_STENCIL_BUFFER_BIT;
	glClearStencil(dat->dwFillStencil);
    }
    if (dat->dwFlags & D3DCLEAR_ZBUFFER) {
	mask |= GL_DEPTH_BUFFER_BIT;
	glClearDepth(dat->dvFillDepth);
	glGetBooleanv(GL_DEPTH_WRITEMASK, &ztest);
	glDepthMask(GL_TRUE); /* Enables Z writing to be sure to delete also the Z buffer */
    }

    glClear(mask);

    if (dat->dwFlags & D3DCLEAR_ZBUFFER)
	glDepthMask(ztest);

    return D3D_OK;
}
