/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "winbase.h"
#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

/* #define DEBUG_VERTEX_DATA */

/* One of the vertex shader input register types is D3DVSDT_D3DCOLOR
 * which is in BGRA ordering.  We must convert this to RGBA ordering
 * since this is what openGL uses.  It seems the best way to do it is
 * to apply the appropriate swizzle to the read from the input color 
 * register. However, since I also wrote code to swizzle on the write 
 * to the output color register, I am going to leave it in for now 
 * (it can be enabled changing the following define to OUTPUT_D3DCOLOR_FIX) */
#define INPUTREG_D3DCOLOR_FIX

DEFAULT_DEBUG_CHANNEL(ddraw);

/********* NV_vertex_program defines from GL specs *************/
#ifndef GL_VERTEX_PROGRAM_NV
#define GL_VERTEX_PROGRAM_NV              0x8620
#define GL_VERTEX_STATE_PROGRAM_NV        0x8621
#define GL_ATTRIB_ARRAY_SIZE_NV           0x8623
#define GL_ATTRIB_ARRAY_STRIDE_NV         0x8624
#define GL_ATTRIB_ARRAY_TYPE_NV           0x8625
#define GL_CURRENT_ATTRIB_NV              0x8626
#define GL_PROGRAM_LENGTH_NV              0x8627
#define GL_PROGRAM_STRING_NV              0x8628
#define GL_MODELVIEW_PROJECTION_NV        0x8629
#define GL_IDENTITY_NV                    0x862A
#define GL_INVERSE_NV                     0x862B
#define GL_TRANSPOSE_NV                   0x862C
#define GL_INVERSE_TRANSPOSE_NV           0x862D
#define GL_MAX_TRACK_MATRIX_STACK_DEPTH_NV 0x862E
#define GL_MAX_TRACK_MATRICES_NV          0x862F
#define GL_MATRIX0_NV                     0x8630
#define GL_MATRIX1_NV                     0x8631
#define GL_MATRIX2_NV                     0x8632
#define GL_MATRIX3_NV                     0x8633
#define GL_MATRIX4_NV                     0x8634
#define GL_MATRIX5_NV                     0x8635
#define GL_MATRIX6_NV                     0x8636
#define GL_MATRIX7_NV                     0x8637
#define GL_CURRENT_MATRIX_STACK_DEPTH_NV  0x8640
#define GL_CURRENT_MATRIX_NV              0x8641
#define GL_VERTEX_PROGRAM_POINT_SIZE_NV   0x8642
#define GL_VERTEX_PROGRAM_TWO_SIDE_NV     0x8643
#define GL_PROGRAM_PARAMETER_NV           0x8644
#define GL_ATTRIB_ARRAY_POINTER_NV        0x8645
#define GL_PROGRAM_TARGET_NV              0x8646
#define GL_PROGRAM_RESIDENT_NV            0x8647
#define GL_TRACK_MATRIX_NV                0x8648
#define GL_TRACK_MATRIX_TRANSFORM_NV      0x8649
#define GL_VERTEX_PROGRAM_BINDING_NV      0x864A
#define GL_PROGRAM_ERROR_POSITION_NV      0x864B
#define GL_VERTEX_ATTRIB_ARRAY0_NV        0x8650
#define GL_VERTEX_ATTRIB_ARRAY1_NV        0x8651
#define GL_VERTEX_ATTRIB_ARRAY2_NV        0x8652
#define GL_VERTEX_ATTRIB_ARRAY3_NV        0x8653
#define GL_VERTEX_ATTRIB_ARRAY4_NV        0x8654
#define GL_VERTEX_ATTRIB_ARRAY5_NV        0x8655
#define GL_VERTEX_ATTRIB_ARRAY6_NV        0x8656
#define GL_VERTEX_ATTRIB_ARRAY7_NV        0x8657
#define GL_VERTEX_ATTRIB_ARRAY8_NV        0x8658
#define GL_VERTEX_ATTRIB_ARRAY9_NV        0x8659
#define GL_VERTEX_ATTRIB_ARRAY10_NV       0x865A
#define GL_VERTEX_ATTRIB_ARRAY11_NV       0x865B
#define GL_VERTEX_ATTRIB_ARRAY12_NV       0x865C
#define GL_VERTEX_ATTRIB_ARRAY13_NV       0x865D
#define GL_VERTEX_ATTRIB_ARRAY14_NV       0x865E
#define GL_VERTEX_ATTRIB_ARRAY15_NV       0x865F
#endif


/******* end NV_vertex_program defines *************************/

#define NV_VP_1_0   0
#define NV_VP_1_1   1

#define SWIZZLE_NONE 0xE4
#define SWIZZLE_X(x) ((x) & 0x03)
#define SWIZZLE_Y(x) (((x) & 0x0C) >> 2)
#define SWIZZLE_Z(x) (((x) & 0x30) >> 4)
#define SWIZZLE_W(x) (((x) & 0xC0) >> 6)

#define GETREGTYPE(x) ((x) & D3DSP_REGTYPE_MASK)
#define GETREGNUM(x) ((x) & D3DSP_REGNUM_MASK)
#define SETREGNUM(x,y) (((x) & ~D3DSP_REGNUM_MASK) | (y))
#define GETSWIZZLE(x) (((x) & D3DVS_SWIZZLE_MASK) >> D3DVS_SWIZZLE_SHIFT)
#define ISADDRRELATIVE(x) ((x) & D3DVS_ADDRMODE_RELATIVE)
#define ISSRCNEGATIVE(x) (((x) & D3DSP_SRCMOD_MASK) == D3DSPSM_NEG)
#define ISWRITEMASKALL(x) (((x) & D3DSP_WRITEMASK_ALL) == D3DSP_WRITEMASK_ALL)
#define ISWRITEMASKX(x) ((x) & D3DSP_WRITEMASK_0)
#define ISWRITEMASKY(x) ((x) & D3DSP_WRITEMASK_1)
#define ISWRITEMASKZ(x) ((x) & D3DSP_WRITEMASK_2)
#define ISWRITEMASKW(x) ((x) & D3DSP_WRITEMASK_3)

#define GETTOKENTYPE(tok) (((tok) & D3DVSD_TOKENTYPEMASK) >> D3DVSD_TOKENTYPESHIFT)
#define GETCONSTCOUNT(tok) (((tok) & D3DVSD_CONSTCOUNTMASK) >> D3DVSD_CONSTCOUNTSHIFT)
#define GETCONSTADDR(tok) (((tok) & D3DVSD_CONSTADDRESSMASK) >> D3DVSD_CONSTADDRESSSHIFT)
#define GETSTREAMNUM(tok) (((tok) & D3DVSD_STREAMNUMBERMASK) >> D3DVSD_STREAMNUMBERSHIFT)
#define GETVERTEXREG(tok) (((tok) & D3DVSD_VERTEXREGMASK) >> D3DVSD_VERTEXREGSHIFT)
#define GETDATATYPE(tok) (((tok) & D3DVSD_DATATYPEMASK) >> D3DVSD_DATATYPESHIFT)
#define ISSTREAMTESS(tok) ((tok) & D3DVSD_STREAMTESSMASK)
#define ISSTREAMDATASKIP(tok) ((tok) & D3DVSD_DATALOADTYPEMASK)
#define GETSKIPCOUNT(tok) (((tok) & D3DVSD_SKIPCOUNTMASK) >> D3DVSD_SKIPCOUNTSHIFT)

#define MAX_VERTEX_ATTRIBUTES   16
#define MAX_INSTRUCTIONS       128
#define MAX_INSTRUCTION_LEN     40  /* this is rather arbitrary - for guestimating size */


VertexShader *gVs = NULL;

static char *VertexProgram;
static int VPLenCur;
static int VPLenMax;

/* VPConcat
 * Concatenate a string onto the end of the current vertex program, 
 * growing the vertex program if necessary.
 */
inline void VPConcat(const char *str) {
  int len = strlen(str);
  /* grow buffer if necessary */
  while( VPLenCur + len > VPLenMax ) {
    VPLenMax = 2*VPLenMax;
    VertexProgram = HeapReAlloc(GetProcessHeap(), 0, VertexProgram, VPLenMax);
  }

  strncat(VertexProgram, str, len);
  VPLenCur += len;
}

/* VPConcatc
 * Concatenate a char onto the end of the current vertex program, 
 * growing the vertex program if necessary
 */
inline void VPConcatc(const char c) {
  /* grow buffer if necessary */
  while( VPLenCur + 1 > VPLenMax ) {
    VPLenMax = 2*VPLenMax;
    VertexProgram = HeapReAlloc(GetProcessHeap(), 0, VertexProgram, VPLenMax);
  }

  VertexProgram[VPLenCur++] = c;
}
  
/* get_vs_token_len
 * Get the length of the current DX8 vertex shader declaration token
 */
static UINT get_vs_token_len(DWORD tok)
{
  UINT len = 1;  // most tokens have a length of 1 DWORD
  
  switch (GETTOKENTYPE(tok)) {
    /* multiword tokens */
  case D3DVSD_TOKEN_CONSTMEM: 
    len += 4*GETCONSTCOUNT(tok);
    break;
  case D3DVSD_TOKEN_EXT:
    len += (tok & D3DVSD_EXTCOUNTMASK) >> D3DVSD_EXTCOUNTSHIFT;
    break;

    /* single word tokens -- fall-through intentional! */
  case D3DVSD_TOKEN_NOP:  /* NOP Token */
  case D3DVSD_TOKEN_STREAM: /* Stream Selector */
  case D3DVSD_TOKEN_STREAMDATA: /* Stream Data Definition */
  case D3DVSD_TOKEN_TESSELLATOR: /* Stream Data Definition (Tesselator data) */
  case D3DVSD_TOKEN_END:
  default:
    break;
  }
  return len;
}

/* find_vs_constmem 
 * returns a pointer to the next CONSTMEM token in the vertex shader
 * declaration block, or NULL if the end of the declaration block is reached
 */
static const DWORD *find_vs_constmem(const DWORD *pDecl)
{
  DWORD tok;

  while( (tok = GETTOKENTYPE(*pDecl)) != D3DVSD_TOKEN_END ) {
    if( tok == D3DVSD_TOKEN_CONSTMEM ) {
      return pDecl;
    } 
    pDecl += get_vs_token_len(*pDecl);
  }
  return NULL;
}

/* find_vs_stream
 * returns a pointer to the next STREAM token in the vertex shader
 * declaration block, or NULL if the end of the declaration block is reached
 */
static const DWORD *find_vs_stream(const DWORD *pDecl)
{
  DWORD tok;

  while( (tok = GETTOKENTYPE(*pDecl)) != D3DVSD_TOKEN_END ) {
    if( tok == D3DVSD_TOKEN_STREAM ) {
      return pDecl;
    } 
    pDecl += get_vs_token_len(*pDecl);
  }
  return NULL;
}

/* get_datatype_dimensionality
 * returns the number of input DWORDS that are occupied by each 
 * datatype
 */
static UINT get_datatype_dimensionality(DWORD datatype) {
  static const BYTE data_dims[] = {
    [D3DVSDT_FLOAT1] = 1,
    [D3DVSDT_FLOAT2] = 2,
    [D3DVSDT_FLOAT3] = 3,
    [D3DVSDT_FLOAT4] = 4,
    [D3DVSDT_D3DCOLOR] = 1,
    [D3DVSDT_UBYTE4] = 1,
    [D3DVSDT_SHORT2] = 2,
    [D3DVSDT_SHORT4] = 4,
  };
  return data_dims[datatype];
}

static const char* const get_datatype(DWORD datatype) {
  static const char* const data_types[] = {
    "D3DVSDT_FLOAT1",
    "D3DVSDT_FLOAT2",
    "D3DVSDT_FLOAT3",
    "D3DVSDT_FLOAT4",
    "D3DVSDT_D3DCOLOR",
    "D3DVSDT_UBYTE4",
    "D3DVSDT_SHORT2",
    "D3DVSDT_SHORT4",
  };
  return data_types[datatype];
}

static UINT dump_vs_declaration(CONST DWORD* pDecl)
{
  CONST DWORD tok = *pDecl;
  UINT len = 1;  /* most tokens have a length of 1 DWORD */
  
  switch ((tok & D3DVSD_TOKENTYPEMASK) >> D3DVSD_TOKENTYPESHIFT) {
  case D3DVSD_TOKEN_NOP:  /* NOP Token */
    /* NOP must be all zeros */
    if( tok == D3DVSD_NOP() ) {
      DPRINTF(" declaration: %08lx NOP()\n", tok);
    } else {
      DPRINTF(" declaration: %08lx invalid NOP token\n", tok);
    }
    break;
  case D3DVSD_TOKEN_STREAM: /* Stream Selector */
    if ((tok & ~(D3DVSD_TOKENTYPEMASK|D3DVSD_STREAMTESSMASK|D3DVSD_STREAMNUMBERMASK)) != 0) {
      /* unused bits are non-zero */
      DPRINTF(" declaration: %08lx invalid STREAM token\n", tok);
    } else if (tok & D3DVSD_STREAMTESSMASK) {
      /* Stream Selector: tesselator stream */
      DPRINTF(" declaration: %08lx STREAM Selector: STREAM(TESS)\n", tok);
    } else {
      /* Stream Selector: stream # */
      DPRINTF(" declaration: %08lx STREAM Selector: STREAM(%ld)\n", tok,
              ((tok & D3DVSD_STREAMNUMBERMASK) >> D3DVSD_STREAMNUMBERSHIFT));
    }
    break;
  case D3DVSD_TOKEN_STREAMDATA: /* Stream Data Definition */
    if (tok & D3DVSD_DATALOADTYPEMASK) {
      if ((tok & ~(D3DVSD_TOKENTYPEMASK|D3DVSD_DATALOADTYPEMASK|D3DVSD_SKIPCOUNTMASK)) != 0) {
        /* unused bits are non-zero */
        DPRINTF(" declaration: %08lx invalid STREAM Data Def (SKIP) token\n", tok);
      } else {
        /* Data Skip: # of dwords to skip */
        DPRINTF(" declaration: %08lx STREAM Data Def: SKIP(%ld)\n", tok,
                ((tok & D3DVSD_SKIPCOUNTMASK) >> D3DVSD_SKIPCOUNTSHIFT));
      }
    }
    else {
      if ((tok & ~(D3DVSD_TOKENTYPEMASK|D3DVSD_DATATYPEMASK|D3DVSD_VERTEXREGMASK)) != 0) {
        /* unused bits are non-zero */
        DPRINTF(" declaration: %08lx invalid STREAM Data Def (REG)\n", tok);
      } else {
        /* Vertex Input Register Load: data type/dimension, vertex register # */
        DPRINTF(" declaration: %08lx STREAM Data Def: REG(%ld,%s)\n", tok,
                ((tok & D3DVSD_VERTEXREGMASK) >> D3DVSD_VERTEXREGSHIFT),
                get_datatype((tok & D3DVSD_DATATYPEMASK) >> D3DVSD_DATATYPESHIFT));
      }
    }
    break;
  case D3DVSD_TOKEN_TESSELLATOR: /* Stream Data Definition (Tesselator data) */
    if (tok & D3DVSD_DATALOADTYPEMASK) {
      if ((tok & ~(D3DVSD_TOKENTYPEMASK|D3DVSD_DATALOADTYPEMASK|
                   D3DVSD_DATATYPEMASK|D3DVSD_VERTEXREGMASK)) != 0 ) {
        DPRINTF(" declaration: %08lx invalid STREAM Data Def (TESSUV) token\n", tok);
      } else {
        /* Tesselator UV data: data type/dimension, Vertex reg out */
        DPRINTF(" declaration: %08lx STREAM Data Def: TESSUV(%s,%ld)\n", tok,
		get_datatype((tok & D3DVSD_DATATYPEMASK) >> D3DVSD_DATATYPESHIFT),
		((tok & D3DVSD_VERTEXREGMASK) >> D3DVSD_VERTEXREGSHIFT));
      }
    } else {
      if ((tok & ~(D3DVSD_TOKENTYPEMASK|D3DVSD_VERTEXREGINMASK|
		   D3DVSD_DATATYPEMASK|D3DVSD_VERTEXREGMASK)) != 0 ) {
	DPRINTF(" declaration: %08lx invalid STREAM Data Def (TESSNORMAL) token\n", tok);
      } else {
        /* Tesselator normal data: Vertex reg in, data type/dimension, Vertex reg out */
	DPRINTF(" declaration: %08lx STREAM Data Def: TESSNORMAL(%ld,%ld,%s)\n", tok,
		((tok & D3DVSD_VERTEXREGINMASK) >> D3DVSD_VERTEXREGINSHIFT),
                ((tok & D3DVSD_VERTEXREGMASK) >> D3DVSD_VERTEXREGSHIFT),
                get_datatype((tok & D3DVSD_DATATYPEMASK) >> D3DVSD_DATATYPESHIFT));
      }
    }
    break;
  case D3DVSD_TOKEN_CONSTMEM:  /* Constant Memory from Shader (multiple DWORD token) */
    if ((tok & ~(D3DVSD_TOKENTYPEMASK|D3DVSD_CONSTCOUNTMASK|D3DVSD_CONSTADDRESSMASK)) != 0 ) {
      DPRINTF(" declaration: %08lx invalid CONSTMEM token\n", tok);
    } else {
      DWORD i;
      DWORD count = (tok & D3DVSD_CONSTCOUNTMASK) >> D3DVSD_CONSTCOUNTSHIFT;
      DWORD addr  = (tok & D3DVSD_CONSTADDRESSMASK) >> D3DVSD_CONSTADDRESSSHIFT;
      DPRINTF(" declaration: %08lx CONSTMEM(%ld,%ld)\n", tok, count, addr);
      pDecl++;
      for( i=0; i<count; i++ ) {
	DPRINTF(" declaration:     c[%ld] = (%08lx,%08lx,%08lx,%08lx)\n", 
		addr, *pDecl, *(pDecl+1), *(pDecl+2), *(pDecl+3));
	pDecl += 4; 
	addr++;
      }
      len += 4*count;
    }
    break;
  case D3DVSD_TOKEN_EXT:
    {
      DWORD i;
      DWORD count = (tok & D3DVSD_EXTCOUNTMASK) >> D3DVSD_EXTCOUNTSHIFT;
      DWORD info = (tok & D3DVSD_EXTINFOMASK) >> D3DVSD_EXTINFOSHIFT;
      DPRINTF(" declaration: %08lx EXTENSION(%ld,%ld)\n", tok, count, info);
      for( i=0; i<count; i++ ) {
	pDecl++;
	DPRINTF(" declaration:     %08lx\n", *pDecl);
      }
      len += count;
    }
    break;
  case D3DVSD_TOKEN_END:
    if( tok == D3DVSD_END() ) {
      DPRINTF(" declaration: %08lx END()\n", tok);
    } else {
      DPRINTF(" declaration: %08lx invalid END token\n", tok);
    }
    break;
  default:
    DPRINTF(" declaration: %08lx (unknown)\n", tok);
    break;
  }
  return len;
}


static DWORD dump_unrecognised_instruction(const DWORD* program)
{
    DWORD length = 1;

    if (TRACE_ON(ddraw)) DPRINTF("unrecognized ins: %08lX\n", *program++);

    while (*program & 0x80000000)
    {
	if (TRACE_ON(ddraw)) DPRINTF("unrecognized ins: %08lX\n", *program++);
	length++;
    }

    /*if (TRACE_ON(ddraw)) DPRINTF("dumped %d unrecognized instructions\n", length);*/
    return length;
}

static BOOL is_comment(DWORD opcode) {
    return ((opcode & D3DSI_OPCODE_MASK) == D3DSIO_COMMENT );
}

static DWORD skip_comment(const DWORD* program)
{
    DWORD len = (*program & D3DSI_COMMENTSIZE_MASK) >> D3DSI_COMMENTSIZE_SHIFT;

    /* assumes the comment is null terminated
       TRACE("COMMENT(%ld): '%s'\n", len, (char *)(program+1)); */

    return (len + 1); /* include the comment opcode itself */
}

static BOOL is_recognized_opcode(DWORD opcode)
{
    static const BYTE opcodes[] =
	{
	    [D3DSIO_NOP] = TRUE,
	    [D3DSIO_MOV] = TRUE,
	    [D3DSIO_ADD] = TRUE,
	    [D3DSIO_SUB] = TRUE,
	    [D3DSIO_MAD] = TRUE,
	    [D3DSIO_MUL] = TRUE,
	    [D3DSIO_RCP] = TRUE,
	    [D3DSIO_RSQ] = TRUE,
	    [D3DSIO_DP3] = TRUE,
	    [D3DSIO_DP4] = TRUE,
	    [D3DSIO_MIN] = TRUE,
	    [D3DSIO_MAX] = TRUE,
	    [D3DSIO_SLT] = TRUE,
	    [D3DSIO_SGE] = TRUE,
	    [D3DSIO_EXP] = TRUE,
	    [D3DSIO_LOG] = TRUE,
	    [D3DSIO_LIT] = TRUE,
	    [D3DSIO_DST] = TRUE,
	    [D3DSIO_FRC] = TRUE,
	    [D3DSIO_M4x4] = TRUE,
	    [D3DSIO_M4x3] = TRUE,
	    [D3DSIO_M3x4] = TRUE,
	    [D3DSIO_M3x3] = TRUE,
	    [D3DSIO_M3x2] = TRUE,
	    [D3DSIO_EXPP] = TRUE,
	    [D3DSIO_LOGP] = TRUE
	};

    if (opcode >= sizeof(opcodes)) return FALSE;

    return opcodes[opcode];
}

static void dump_opcode_nv(DWORD opcode)
{
    static const char * const opcode_names[] =
	{
	    [D3DSIO_MOV] = "MOV ",
	    [D3DSIO_ADD] = "ADD ",
	    [D3DSIO_SUB] = "SUB ",
	    [D3DSIO_MAD] = "MAD ",
	    [D3DSIO_MUL] = "MUL ",
	    [D3DSIO_RCP] = "RCP ",
	    [D3DSIO_RSQ] = "RSQ ",
	    [D3DSIO_DP3] = "DP3 ",
	    [D3DSIO_DP4] = "DP4 ",
	    [D3DSIO_MIN] = "MIN ",
	    [D3DSIO_MAX] = "MAX ",
	    [D3DSIO_SLT] = "SLT ",
	    [D3DSIO_SGE] = "SGE ",
	    [D3DSIO_EXP] = "EXP ",
	    [D3DSIO_LOG] = "LOG ",
	    [D3DSIO_LIT] = "LIT ",
	    [D3DSIO_DST] = "DST ",
	    [D3DSIO_EXPP] = "EXP ",
	    [D3DSIO_LOGP] = "LOG "
	};

    VPConcat(opcode_names[opcode]);
}

static DWORD num_opcode_operands(DWORD opcode)
{
    static const BYTE num_params[] =
	{
	    [D3DSIO_NOP] = 0,
	    [D3DSIO_MOV] = 2,
	    [D3DSIO_ADD] = 3,
	    [D3DSIO_SUB] = 3,
	    [D3DSIO_MAD] = 4,
	    [D3DSIO_MUL] = 3,
	    [D3DSIO_RCP] = 2,
	    [D3DSIO_RSQ] = 2,
	    [D3DSIO_DP3] = 3,
	    [D3DSIO_DP4] = 3,
	    [D3DSIO_MIN] = 3,
	    [D3DSIO_MAX] = 3,
	    [D3DSIO_SLT] = 3,
	    [D3DSIO_SGE] = 3,
	    [D3DSIO_EXP] = 2,
	    [D3DSIO_LOG] = 2,
	    [D3DSIO_LIT] = 2,
	    [D3DSIO_DST] = 3,
	    [D3DSIO_FRC] = 2,
	    [D3DSIO_M4x4] = 3,
	    [D3DSIO_M4x3] = 3,
	    [D3DSIO_M3x4] = 3,
	    [D3DSIO_M3x3] = 3,
	    [D3DSIO_M3x2] = 3,
	    [D3DSIO_EXPP] = 2,
	    [D3DSIO_LOGP] = 2
	};

    return num_params[opcode];
}

static DWORD increment_register(DWORD reg) {
  DWORD newreg;
  DWORD regnum = GETREGNUM(reg);
  regnum++;
  newreg = SETREGNUM(reg, regnum);
  return newreg;
}

static void dump_register_nv(DWORD reg)
{
  static const char * const rastout_names[] = { "o[HPOS]", "o[FOGC]", "o[PSIZ]" };
  static const char * const attrout_names[] = { "o[COL0]", "o[COL1]" };
  char outbuf[20] = "\0";
  
  DWORD regnum = GETREGNUM(reg);
  
  switch (GETREGTYPE(reg)) {
  case D3DSPR_TEMP:
    sprintf(outbuf, "R%lu", regnum);
    break;
    
  case D3DSPR_INPUT:
    sprintf(outbuf, "v[%lu]", regnum);
    break;
    
  case D3DSPR_CONST:
    /* XXX What is the correct interpretation of regnum when a relative
     * offset is used?
     * GL_NV_vertex_program uses the range -64 to 63. 
     * (and how are negative values handled??)
     */
    sprintf(outbuf, "c[%s%lu]",
	    ISADDRRELATIVE(reg) ? "A0.x+" : "",
	    regnum);
    break;
    
  case D3DSPR_ADDR:
    sprintf(outbuf, "A%lu", regnum);
    break;
    
  case D3DSPR_RASTOUT:
    sprintf(outbuf, "%s", rastout_names[regnum]);
    break;
    
  case D3DSPR_ATTROUT:
    sprintf(outbuf, "%s", attrout_names[regnum]);
    break;
    
  case D3DSPR_TEXCRDOUT:
    sprintf(outbuf, "o[TEX%lu]", regnum);
    break;

  default:
    TRACE("unknown register\n");
  }

  VPConcat(outbuf);
}

static void dump_writemask_nv(DWORD param)
{
    if( !ISWRITEMASKALL(param) )
    {
	VPConcat(".");

	if (ISWRITEMASKX(param)) VPConcat("x");
	if (ISWRITEMASKY(param)) VPConcat("y");
	if (ISWRITEMASKZ(param)) VPConcat("z");
	if (ISWRITEMASKW(param)) VPConcat("w");
    }
}

static void dump_output_parameter_nv(DWORD param)
{
    dump_register_nv(param);
    dump_writemask_nv(param);
}

static void dump_swizzle_nv(DWORD param)
{
    static const char components[] = "xyzw";
    char outbuf[10] = "\0";

    DWORD swizzle = GETSWIZZLE(param);

    if (swizzle == SWIZZLE_NONE) return;

    if (SWIZZLE_X(swizzle) == SWIZZLE_Y(swizzle)
	&& SWIZZLE_X(swizzle) == SWIZZLE_Z(swizzle)
	&& SWIZZLE_X(swizzle) == SWIZZLE_W(swizzle))
    {
	sprintf(outbuf, ".%c", components[SWIZZLE_X(swizzle)]);
    }
    else
    {
	sprintf(outbuf, ".%c%c%c%c",
		components[SWIZZLE_X(swizzle)],
		components[SWIZZLE_Y(swizzle)],
		components[SWIZZLE_Z(swizzle)],
		components[SWIZZLE_W(swizzle)]);
    }
    VPConcat(outbuf);
}

#ifdef INPUTREG_D3DCOLOR_FIX
/* if the input register is of type D3DVSDT_D3DCOLOR, then it is 
 * in the order BGRA and we need it to be in RGBA (ie swap the x,z components)
 */
static void dump_swizzle_inputcolorfix_nv(DWORD param)
{
    static const char components_swap[] = "zyxw";
    char outbuf[10] = "\0";

    DWORD swizzle = GETSWIZZLE(param);

    if (SWIZZLE_X(swizzle) == SWIZZLE_Y(swizzle)
	&& SWIZZLE_X(swizzle) == SWIZZLE_Z(swizzle)
	&& SWIZZLE_X(swizzle) == SWIZZLE_W(swizzle))
    {
	sprintf(outbuf, ".%c", components_swap[SWIZZLE_X(swizzle)]);
    }
    else
    {
	sprintf(outbuf, ".%c%c%c%c",
		components_swap[SWIZZLE_X(swizzle)],
		components_swap[SWIZZLE_Y(swizzle)],
		components_swap[SWIZZLE_Z(swizzle)],
		components_swap[SWIZZLE_W(swizzle)]);
    }
    VPConcat(outbuf);
}
#endif /* INPUTREG_D3DCOLOR_FIX */

#ifdef OUTPUTREG_D3DCOLOR_FIX
/* if writing the the output color registers, the color
 * values are currently in BGRA order(dx), and the output is expecting them 
 * in RGBA order (gl), so we must reorder them appropriately
 */
static void dump_colorfix_swizzle_nv(DWORD param)
{
    static const char components_fix[] = "zyxw";
    static const char components[] = "xyzw";
    char outbuf[10] = "\0";
    DWORD swizzle = GETSWIZZLE(param);

    if (swizzle == SWIZZLE_NONE) {
	/* reading all the components, this should apply the colorfix */
	TRACE("applying colorfix swizzle to output color register (no swizzle)\n");
	sprintf(outbuf, ".%c%c%c%c",
		components_fix[SWIZZLE_X(swizzle)],
		components_fix[SWIZZLE_Y(swizzle)],
		components_fix[SWIZZLE_Z(swizzle)],
		components_fix[SWIZZLE_W(swizzle)]);
    } 
    else if (SWIZZLE_X(swizzle) == SWIZZLE_Y(swizzle)
	&& SWIZZLE_X(swizzle) == SWIZZLE_Z(swizzle)
	&& SWIZZLE_X(swizzle) == SWIZZLE_W(swizzle))
    {
	/* if only one component is specified, it is likely that it is being 
	 * broadcast to all fields, and thus applying the colorfix would 
	 * probably not the be the correct thing to do */
      	TRACE("not applying colorfix swizzle to output color register\n");
	sprintf(outbuf, ".%c", components[SWIZZLE_X(swizzle)]);
    }
    else
    {
	/* reading multiple components, it is likely that we want to reorder
	 * the color components -- but I haven't actually seen a vertex program 
	 * that uses this case yet */
	TRACE("applying colorfix swizzle to output color register (swizzle)\n");
	sprintf(outbuf, ".%c%c%c%c",
		components_fix[SWIZZLE_X(swizzle)],
		components_fix[SWIZZLE_Y(swizzle)],
		components_fix[SWIZZLE_Z(swizzle)],
		components_fix[SWIZZLE_W(swizzle)]);
    }
    VPConcat(outbuf);
}
#endif /* OUTPUTREG_D3DCOLOR_FIX */ 

static void dump_input_parameter_nv(DWORD param)
{
    if (ISSRCNEGATIVE(param)) {
      VPConcat("-");
    }
    dump_register_nv(param);
#ifdef INPUTREG_D3DCOLOR_FIX
    /* special case if input stream is D3DVSTD_D3DCOLOR */
    if( GETREGTYPE(param) == D3DSPR_INPUT && 
	GET_INPUT_REG_TYPE(gVs->input_reg[GETVERTEXREG(param)].type) == D3DVSDT_D3DCOLOR ) {
      dump_swizzle_inputcolorfix_nv(param);
    } else 
#endif
      dump_swizzle_nv(param);
    
}
static void dump_neg_input_parameter_nv(DWORD param)
{
    if (!ISSRCNEGATIVE(param)) {
      VPConcat("-");
    }
    dump_register_nv(param);
#ifdef INPUTREG_D3DCOLOR_FIX    
    /* special case if input stream is D3DVSTD_D3DCOLOR */
    if( GETREGTYPE(param) == D3DSPR_INPUT && 
	GET_INPUT_REG_TYPE(gVs->input_reg[GETVERTEXREG(param)].type) == D3DVSDT_D3DCOLOR ) {
      dump_swizzle_inputcolorfix_nv(param);
    } else
#endif     
      dump_swizzle_nv(param);
}

static DWORD translate_instruction_nv(const DWORD* program, int version)
{
  static const char components[] = "xyzw";
  if (is_recognized_opcode(*program)) {
    DWORD i, tmp;
    DWORD opcode = *program & D3DSI_OPCODE_MASK;
    DWORD num_params = num_opcode_operands(opcode);
    DWORD dst = program[1];
    const DWORD *src = &(program[2]);

    switch(opcode) {
    case D3DSIO_NOP:  /* no need for this op */
      break; 
    case D3DSIO_MOV:  /* could be MOV or ARL, depending on dest */
      if( GETREGTYPE(dst) == D3DSPR_ADDR ) { 
	VPConcat("ARL "); 
	/* not sure if DX8-vs dst & src mask are implicit or explicit, so check */
	dump_output_parameter_nv(dst);
	if( ISWRITEMASKALL(dst) ) {
	  /* add the explicit .x writemask */
	  VPConcat(".x");
	}
	VPConcat(", ");
	dump_input_parameter_nv(src[0]);
	if( GETSWIZZLE(src[0]) == SWIZZLE_NONE ){
	  /* add the default source substript (.x) */
	  VPConcat(".x");
	} 
	VPConcat(";\n");
      } else {
	dump_opcode_nv(D3DSIO_MOV);
	dump_output_parameter_nv(dst); VPConcat(", "); 
#ifdef OUTPUTREG_D3DCOLOR_FIX	
	/* if writing to color output register we must apply the BGRA->RGBA colorfix */
	if( GETREGTYPE(dst) == D3DSPR_ATTROUT ) {
	  if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	  dump_register_nv(src[0]);
	  dump_colorfix_swizzle_nv(src[0]);
	} else  
#endif	  
	  dump_input_parameter_nv(src[0]); 
	VPConcat(";\n");
      }
      break;

      /* SCALAR ops -- translate directly to NV1.0/1.1
       * but may need to add an explicit source subscript
       * (assume .w if not explicit)  */
    case D3DSIO_RCP:  
    case D3DSIO_RSQ:  
    case D3DSIO_EXPP:
    case D3DSIO_LOGP:
      dump_opcode_nv(opcode);
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_input_parameter_nv(src[0]); 
      if( GETSWIZZLE(src[0]) == SWIZZLE_NONE ) {
	/* add the default source subscript (.w) */
	VPConcat(".w");
      }
      VPConcat(";\n");
      break;
      
      /* these ops translate directly to NV1.0/1.1 */
    case D3DSIO_ADD:
    case D3DSIO_MAD:
    case D3DSIO_MUL:
    case D3DSIO_DP3:
    case D3DSIO_DP4:
    case D3DSIO_MIN:
    case D3DSIO_MAX:
    case D3DSIO_SLT:
    case D3DSIO_SGE:
    case D3DSIO_LIT:
    case D3DSIO_DST: 
      dump_opcode_nv(opcode);
      
      if( num_params > 0 ) dump_output_parameter_nv(dst);
      for( i=0; i<num_params-1; i++ ) {
	VPConcat(", ");
	dump_input_parameter_nv(src[i]);
      }
      VPConcat(";\n");
      break;

    case D3DSIO_SUB:  /* supported on NV1.1, but simulated on NV1.0 */
      if( version == NV_VP_1_1 ) {
	/* SUB dst, src0, src1 */
	dump_opcode_nv(D3DSIO_SUB);
	dump_output_parameter_nv(dst); VPConcat(", "); 
	dump_input_parameter_nv(src[0]); VPConcat(", "); 
	dump_input_parameter_nv(src[1]); VPConcat(";\n");
      } else {
	/* ADD dst, src0, -src1 */
	dump_opcode_nv(D3DSIO_ADD);
	dump_output_parameter_nv(dst); VPConcat(", "); 
	dump_input_parameter_nv(src[0]); VPConcat(", "); 
	dump_neg_input_parameter_nv(src[1]); VPConcat(";\n");
      }
      break;

      /* Matrix transforms macros - implemented using DP3/DP4
	 writemask, negation or swizzle modifiers cannot be used */
    case D3DSIO_M3x2: /* 2D matrix transform */
      VPConcat("# expanding: 'M3x2 "); 
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_input_parameter_nv(src[0]); VPConcat(", ");
      dump_input_parameter_nv(src[1]); VPConcat(";'\n");
      VPConcat("DP3 "); 
      dump_register_nv(dst); VPConcat(".x, ");
      dump_register_nv(src[0]); VPConcat(", ");
      dump_register_nv(src[1]); VPConcat(";\n");
      VPConcat("DP3 ");
      dump_register_nv(dst); VPConcat(".y, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(src[1]);
      dump_register_nv(tmp); VPConcat(";\n");
      break;

    case D3DSIO_M3x3: /* 3D matrix transform - no w */
      VPConcat("# expanding: 'M3x3 ");
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_input_parameter_nv(src[0]); VPConcat(", ");
      dump_input_parameter_nv(src[1]); VPConcat(";'\n");
      VPConcat("DP3 "); 
      dump_register_nv(dst); VPConcat(".x, ");
      dump_register_nv(src[0]); VPConcat(", ");
      dump_register_nv(src[1]); VPConcat(";\n");
      VPConcat("DP3 ");
      dump_register_nv(dst); VPConcat(".y, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(src[1]);
      dump_register_nv(tmp); VPConcat(";\n");
      VPConcat("DP3 ");
      dump_register_nv(dst); VPConcat(".z, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(tmp);
      dump_register_nv(tmp); VPConcat(";\n");
      break;
     
    case D3DSIO_M3x4: /* 3D matrix transform - no translation */
      VPConcat("# expanding: 'M3x4 ");
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_input_parameter_nv(src[0]); VPConcat(", ");
      dump_input_parameter_nv(src[1]); VPConcat(";'\n");
      VPConcat("DP3 "); 
      dump_register_nv(dst); VPConcat(".x, ");
      dump_register_nv(src[0]); VPConcat(", ");
      dump_register_nv(src[1]); VPConcat(";\n");
      VPConcat("DP3 ");
      dump_register_nv(dst); VPConcat(".y, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(src[1]);
      dump_register_nv(tmp); VPConcat(";\n");
      VPConcat("DP3 ");
      dump_register_nv(dst); VPConcat(".z, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(tmp);
      dump_register_nv(tmp); VPConcat(";\n");
      VPConcat("DP3 ");
      dump_register_nv(dst); VPConcat(".w, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(tmp);
      dump_register_nv(tmp); VPConcat(";\n");
      break;
     
    case D3DSIO_M4x3: /* 3D matrix transform - no projective */
      VPConcat("# expanding: 'M4x3 ");
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_input_parameter_nv(src[0]); VPConcat(", ");
      dump_input_parameter_nv(src[1]); VPConcat(";'\n");
      VPConcat("DP4 "); 
      dump_register_nv(dst); VPConcat(".x, ");
      dump_register_nv(src[0]); VPConcat(", ");
      dump_register_nv(src[1]); VPConcat(";\n");
      VPConcat("DP4 ");
      dump_register_nv(dst); VPConcat(".y, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(src[1]);
      dump_register_nv(tmp); VPConcat(";\n");
      VPConcat("DP4 ");
      dump_register_nv(dst); VPConcat(".z, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(tmp);
      dump_register_nv(tmp); VPConcat(";\n");
      break;

    case D3DSIO_M4x4: /* 3D projective tranform */
      VPConcat("# expanding: 'M4x4 ");
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_input_parameter_nv(src[0]); VPConcat(", ");
      dump_input_parameter_nv(src[1]); VPConcat(";'\n");
      VPConcat("DP4 "); 
      dump_register_nv(dst); VPConcat(".x, ");
      dump_register_nv(src[0]); VPConcat(", ");
      dump_register_nv(src[1]); VPConcat(";\n");
      VPConcat("DP4 ");
      dump_register_nv(dst); VPConcat(".y, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(src[1]);
      dump_register_nv(tmp); VPConcat(";\n");
      VPConcat("DP4 ");
      dump_register_nv(dst); VPConcat(".z, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(tmp);
      dump_register_nv(tmp); VPConcat(";\n");
      VPConcat("DP4 ");
      dump_register_nv(dst); VPConcat(".w, ");
      dump_register_nv(src[0]); VPConcat(", ");
      tmp = increment_register(tmp);
      dump_register_nv(tmp); VPConcat(";\n");
      break;
     
    case D3DSIO_FRC: 
      /* FRC dst, src -- computes fraction components of x&y components 
       * emulate using sequence of EXP and MOV
       * (Only the .y and .xy write masks seem to be valid in DX8)
       */
      {
	DWORD swzl = GETSWIZZLE(src[0]); 
	VPConcat("# emulating: 'FRC ");
	dump_output_parameter_nv(dst); VPConcat(", ");
	dump_input_parameter_nv(src[0]); VPConcat(";'\n");
	if( ISWRITEMASKX(dst) && ISWRITEMASKY(dst) ) {
	  int X_is_y = ( components[SWIZZLE_X(swzl)] == 'y' );
	  int src_is_dst = ( GETREGTYPE(src[0]) == GETREGTYPE(dst) &&
			   GETREGNUM(src[0]) == GETREGNUM(dst) );
	  
 	  if( SWIZZLE_X(swzl) == SWIZZLE_Y(swzl) ) { 
	    /* special case: src X,Y are the same -- save an instruction*/
	    VPConcat("EXP ");
	    dump_register_nv(dst); VPConcat(".y, ");
	    if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	    dump_register_nv(src[0]); VPConcat(".");
	    VPConcatc(components[SWIZZLE_X(swzl)]);

	    VPConcat(";\nMOV ");
	    dump_register_nv(dst); VPConcat(".x, ");
	    dump_register_nv(dst); VPConcat(".y;\n");
	    
	  } else if( src_is_dst ) {
	    /* special case: src,dst registers are the same --
	     * must ensure that the y component of the register is used first
	     * and may need an extra swap */
	    VPConcat("EXP "); 
	    dump_register_nv(dst); VPConcat(".y, ");
	    if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	    dump_register_nv(src[0]); VPConcat(".");
	    if( X_is_y ) {
	      VPConcatc(components[SWIZZLE_X(swzl)]);
	    } else {
	      VPConcatc(components[SWIZZLE_Y(swzl)]);
	    }
	    
	    VPConcat(";\nMOV ");
	    dump_register_nv(dst); VPConcat(".xy, ");
	    dump_register_nv(dst); VPConcat(".y");
	    if( X_is_y ) {
	      VPConcatc(components[SWIZZLE_Y(swzl)]);
	    } else {
	      VPConcatc(components[SWIZZLE_X(swzl)]);
	    }
	    
	    VPConcat(";\nEXP ");
	    dump_register_nv(dst); VPConcat(".y, ");
	    if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	    dump_register_nv(dst); VPConcat(".");
	    VPConcatc(components[SWIZZLE_Y(swzl)]); 
	    VPConcat(";\n");
	    if( !X_is_y ) {
	      VPConcat("MOV ");
	      dump_register_nv(dst); VPConcat(".xy, ");
	      dump_register_nv(dst); VPConcat(".yx;\n");
	    }
	    
	  } else {
	    /* general case */
	    VPConcat("EXP "); 
	    dump_register_nv(dst); VPConcat(".y, ");
	    if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	    dump_register_nv(src[0]); VPConcat(".");
	    VPConcatc(components[SWIZZLE_X(swzl)]);
	    
	    VPConcat(";\nMOV ");
	    dump_register_nv(dst); VPConcat(".x, ");
	    dump_register_nv(dst); VPConcat(".y;\n");
	    
	    VPConcat("EXP ");
	    dump_register_nv(dst); VPConcat(".y, ");
	    if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	    dump_register_nv(src[0]); VPConcat(".");
	    VPConcatc(components[SWIZZLE_Y(swzl)]); 
	    VPConcat(";\n");
	  }
	  
	} else if( ISWRITEMASKY(dst) ) {
	  VPConcat("EXP ");
	  dump_register_nv(dst); VPConcat(".y, ");
	  if (ISSRCNEGATIVE(src[0])) VPConcat("-");
	  dump_register_nv(src[0]); VPConcat(".");
	  VPConcatc(components[SWIZZLE_X(swzl)]);
	  VPConcat(";\n");
	}
	VPConcat("# end emulation\n");
      }
      break;

      /* High precision ops -- must be emulated on NV1.0/1.1
       * To properly simulate these instructions requires
       * 2-3 constants, 1 temp register, and 10 instructions, as 
       * explained in the NVIDIA OpenGL Extension specs.
       * Since this would be very difficult, we'll just implement
       * by calling the low precision versions, and replicating 
       * the results as necessary */
    case D3DSIO_EXP: 
    case D3DSIO_LOG:
      WARN("Emulating high precision EXP/LOG--may have unexpected side effects!!\n");
      VPConcat("# Emulating high precision op, with low precision op\n");
      dump_opcode_nv(opcode); 
      dump_register_nv(dst); VPConcat(".z, ");
      dump_input_parameter_nv(src[0]); VPConcat(";\n");
      VPConcat("MOV "); 
      dump_output_parameter_nv(dst); VPConcat(", ");
      dump_register_nv(dst); VPConcat(".z;\n");
      break;
      
    default:
      TRACE("NV instruction %08lx not implemented yet\n", opcode);
    }

    return num_params + 1;

  } else if (is_comment(*program)) {
    return skip_comment(program);

  } else {
    return dump_unrecognised_instruction(program);
  }
}

static BOOL check_program_version_tag(DWORD version)
{
    /* check the version tag */
    if (HIWORD(version) != 0xFFFE)
    {
	WARN("vertex program has no version tag\n");
	return FALSE;
    }

    /* 1.0, 1.1 are supported. */
    if (D3DSHADER_VERSION_MAJOR(version) != 1
	|| D3DSHADER_VERSION_MINOR(version) > 1)
    {
	WARN("unknown vertex program version: %ld.%ld\n",
	     D3DSHADER_VERSION_MAJOR(version),
	     D3DSHADER_VERSION_MINOR(version));
	return FALSE;
    }

    return TRUE;
}


static void GL_VP_create_NV(GL_D3D_priv* priv, VertexShader *vs, int version) {
  GLuint name = 0;
  GLint errpos;
  const DWORD * program = vs->pFunction;

  /* initialize things for the vertex program string */
  VPLenMax = MAX_INSTRUCTIONS * MAX_INSTRUCTION_LEN;
  VPLenCur = 0;
  VertexProgram = HeapAlloc(GetProcessHeap(), 0, VPLenMax*sizeof(char));
  VertexProgram[0] = '\0';

  /* add NV version string */
  switch (version) {
  case NV_VP_1_0: 
    VPConcat("!!VP1.0\n");
    break;
  case NV_VP_1_1:
    VPConcat("!!VP1.1\n");
    break;
  }

  program++;  /* skip the version header */
  gVs = vs;

  while (*program != D3DVS_END()) {
    program += translate_instruction_nv(program, version);
  }
  VPConcat("END\n");
  gVs = NULL;

  if( TRACE_ON(ddraw) ) {
    fprintf(stderr, "NV vertex program string:\n%s", VertexProgram);
  }

  /* generated a name(id) for the program */
  GLD(priv)->glGenProgramsNV(1, &name);
  TRACE("generated program name=%u\n", name);

  /* save the vertex program name(id) for later use */
  vs->DriverReserved0 = (DWORD)name;
  
  /* load the vertex program and check for errors */
  GLD(priv)->glLoadProgramNV(GL_VERTEX_PROGRAM_NV, name, 
			     strlen(VertexProgram), (GLubyte *)VertexProgram);

  glGetIntegerv(GL_PROGRAM_ERROR_POSITION_NV, &errpos);
  if( errpos < 0 ) {
    TRACE("Vertex Program (NV) (name=%u) loaded successfully\n", name);
  } else if( errpos < strlen(VertexProgram) ) {
    WARN("syntax error in Vertex Program (NV) at offset=%d\n", errpos);
    fprintf(stderr, "%s", &VertexProgram[errpos]);
  } else {
    WARN("semantic error in Vertex Program (NV) errpos=%d\n", errpos);
  }

  /* delete the internal vertex program */
  HeapFree(GetProcessHeap(), 0, VertexProgram);
  
}

/* =================== Vertex Program API ===================== */

static HRESULT 
GL_VP_create(GL_D3D_priv* priv, VertexShader *vs)
{
  DWORD vs_version = vs->pFunction[0];

  /* make sure we have a recognized shader version */
  if( !check_program_version_tag(vs_version) ) {
    TRACE("unrecognized vertex shader version 0x%08lx\n", vs_version);
    TRACE("vs.%ld.%ld\n", D3DSHADER_VERSION_MAJOR(vs_version),
	  D3DSHADER_VERSION_MINOR(vs_version));
    return D3DERR_INVALIDCALL;
  }

  /* check what GL extension we are using */
  if( HAVE_GL_EXT(priv, NV_vertex_program1_1) || HAVE_GL_EXT(priv, NV_vertex_program) ) {
    int version;
    
    if( HAVE_GL_EXT(priv, NV_vertex_program1_1) ) {
      TRACE("using NV_vertex_program1_1\n");
      version = NV_VP_1_1;
    } else {
      TRACE("using NV_vertex_program\n");
      version = NV_VP_1_0;
    }
    
    /* convert DX8VertexShader function block into NV vertex program definition */
    GL_VP_create_NV(priv, vs, version); 
    
#if 0
  } else if( HAVE_GL_EXT(priv, ARB_vertex_program) ) {
    /* when implemented this should be the default one! */
    TRACE("using ARB_vertex_program\n");
    
    /*CreateVertexProgramARB(...); */
    FIXME("ARB_vertex_program not implemented yet!\n");
#endif
  }
  CHECK_GL_ERROR("VP_create");
  return D3D_OK;
}

static void 
GL_VP_set_constant(GL_D3D_priv* priv, GLuint reg, GLuint count, float *consts) 
{
  /* update the cached copy of the constants */
  memcpy(priv->d3dp->vertex_const + 4*reg, consts, 4*sizeof(float)*count);

  if(HAVE_GL_EXT(priv, NV_vertex_program) ) {
#if 0    
    /* this code block is for debugging -- writes the registers one at a time 
     * and checks the values in the register after the write */
    GLuint i;
    TRACE("NV: setting vertex program constants reg=%d, count=%d\n", reg, count);
    for( i=0; i<count; i++ ) {
      GLfloat set[4];
      TRACE("  set reg=%d, (%f,%f,%f,%f)\n", reg+i, 
	  consts[i*4+0], consts[i*4+1], consts[i*4+2], consts[i*4+3]);
      GLD(priv)->glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV, reg+i,
	  consts[i*4+0], consts[i*4+1], consts[i*4+2], consts[i*4+3]);

      /* read back the values from the registers to see if they were set properly */
      GLD(priv)->glGetProgramParameterfvNV(GL_VERTEX_PROGRAM_NV, reg+i, 
	                                 GL_PROGRAM_PARAMETER_NV, set);
      TRACE("  got reg=%d, (%f,%f,%f,%f)\n", reg+i, 
	  set[0], set[1], set[2], set[3]);
    }
#else
    TRACE("NV: setting vertex program constants reg=%d, count=%d\n", reg, count);
    GLD(priv)->glProgramParameters4fvNV(GL_VERTEX_PROGRAM_NV, 
					reg, count, consts); 
#endif
  }
  CHECK_GL_ERROR("VP_set_constant");
}

static void 
GL_VP_enable(GL_D3D_priv* priv, GLuint name) 
{
  if( HAVE_GL_EXT(priv, NV_vertex_program) ) {
    TRACE("NV: enabling vertex programs and setting to name=%u\n", name);
    
    GLD(priv)->glBindProgramNV(GL_VERTEX_PROGRAM_NV, name);
    if (!priv->isVertexProgramEnabled) {
      glEnable(GL_VERTEX_PROGRAM_NV);
      priv->isVertexProgramEnabled = TRUE;
    }
  }
  CHECK_GL_ERROR("VP_enable");
}

static void 
GL_VP_disable(GL_D3D_priv* priv)
{
  if( HAVE_GL_EXT(priv, NV_vertex_program) ) {
    if (priv->isVertexProgramEnabled) {
      TRACE("NV: disabling vertex program\n");
      glDisable(GL_VERTEX_PROGRAM_NV);
      priv->isVertexProgramEnabled = FALSE;
    }
  }
  CHECK_GL_ERROR("VP_disable");
}

static void 
GL_VP_delete(GL_D3D_priv *priv, GLuint name) 
{
  if(HAVE_GL_EXT(priv, NV_vertex_program) ) {
    TRACE("deleting vertex program (NV) name=%u\n", name);
    
    GLD(priv)->glDeleteProgramsNV(1, &name);      
  }
  CHECK_GL_ERROR("VP_delete");
}


/* ===============  Externally called functions ========================== */

HRESULT D3D_GL_create_vertex_shader(GL_D3D_priv* priv, 
				    LPD3DHAL_DP2COMMAND cmd, 
				    LPVOID data1, 
				    LPVOID data2)
{
  D3DHAL_DP2CREATEVERTEXSHADER *dat = (D3DHAL_DP2CREATEVERTEXSHADER *)data1;
  VertexShader *vs = VS_FROM_HANDLE(dat->dwHandle);
  
  TRACE("Create Vertex Shader handle=0x%lx\n", dat->dwHandle);

  if( IS_VS(dat->dwHandle) ) {

    /* handle the declaration block */
    if( vs->pDeclaration ) {
      CONST DWORD *decl = vs->pDeclaration;

      if( TRACE_ON(ddraw) ) {
        CONST DWORD *ptr = vs->pDeclaration;
        while (*ptr != D3DVSD_END()) {
          ptr += dump_vs_declaration(ptr);
        }
      }
      
      TRACE("parsing vertex shader declaration block\n");
      /* get the stream defs from the declaration block */
      while( (decl=find_vs_stream(decl)) != NULL ) {
	if( ISSTREAMTESS(*decl) ) {
	  FIXME("Stream(TESS) is not currently implemented\n");
	  decl++;
	} else {
	  BOOL done = FALSE;
	  BYTE stream = GETSTREAMNUM(*decl);
	  BYTE offset = 0;
	  
	  decl++;
	  do {
	    switch (GETTOKENTYPE(*decl)) {
	    case D3DVSD_TOKEN_TESSELLATOR:
	      FIXME("Tesselator data -- not currently implemented\n");
	      /* intentional fall-through */
	    case D3DVSD_TOKEN_NOP:
	      decl++;
	      break;
	    case D3DVSD_TOKEN_STREAMDATA:
	      if( ISSTREAMDATASKIP(*decl) ) {
		offset += GETSKIPCOUNT(*decl)*sizeof(DWORD);
	      } else {
		BYTE reg = GETVERTEXREG(*decl);
		BYTE type = GETDATATYPE(*decl);
		vs->input_reg[reg].stream = stream;
		vs->input_reg[reg].offset = offset;
		vs->input_reg[reg].type = SET_INPUT_REG_TYPE(type);
		TRACE("input_reg[%d]=[str(%d),off(%d),type(%s)]\n", 
		      reg, vs->input_reg[reg].stream, 
		      vs->input_reg[reg].offset, 
		      get_datatype(GET_INPUT_REG_TYPE(vs->input_reg[reg].type)));
		offset += get_datatype_dimensionality(type)*sizeof(DWORD);
	      }
	      decl++;
	      break;
	    default:
	      done = TRUE;
	      break;
	    }
	  } while (!done);
	}  
      }  
    } /* end declaration block */
    
    /* handle the function block */
    if( vs->pFunction ) {
      TRACE("parsing vertex shader function block\n");
      return GL_VP_create(priv, vs);
    }
  }

  return D3D_OK;
}

HRESULT D3D_GL_set_vertex_shader(GL_D3D_priv* priv, 
				 LPD3DHAL_DP2COMMAND cmd, 
				 LPVOID data1, 
				 LPVOID data2)
{
  D3DHAL_DP2VERTEXSHADER *dat = (D3DHAL_DP2VERTEXSHADER *)data1;
  VertexShader *vs = VS_FROM_HANDLE(dat->dwHandle);
  TRACE("(%p)->(0x%lx)\n", priv, dat->dwHandle);

  priv->glFlags &= ~(GLF_VERTEX_DECLARATION|GLF_VERTEX_PROGRAM);
  priv->glFlags |= GLF_NEW_TNL_MODE;
  if( IS_VS(dat->dwHandle) ) {

    priv->current_vs = vs;

    if( vs->pDeclaration ) {
      CONST DWORD* decl = vs->pDeclaration; 
      TRACE("setting vertex shader declaration\n");
      priv->glFlags |= GLF_VERTEX_DECLARATION;

      while( (decl=find_vs_constmem(decl)) != NULL ) {
	GLuint count = GETCONSTCOUNT(*decl);
	GLuint addr = GETCONSTADDR(*decl);

	decl++;
	GL_VP_set_constant(priv, addr, count, (float *)decl);
	
	decl += 4*count;
      }
    }

    if( vs->pFunction ) {
      TRACE("setting vertex shader function\n");
      priv->glFlags |= GLF_VERTEX_PROGRAM;
      GL_VP_enable(priv, (GLuint)vs->DriverReserved0);
    } else {
      GL_VP_disable(priv);
    }

  } else if (priv->current_vs) {
    priv->current_vs = NULL;
    GL_VP_disable(priv);
  }
  
  return D3D_OK;
}

HRESULT D3D_GL_set_vertex_shader_const(GL_D3D_priv* priv, 
				       LPD3DHAL_DP2COMMAND cmd, 
				       LPVOID data1, 
				       LPVOID data2)
{
  D3DHAL_DP2SETVERTEXSHADERCONST *dat = (D3DHAL_DP2SETVERTEXSHADERCONST *)data1;
  
  TRACE("(%p)->(%ld,%p,%ld)\n", priv, dat->dwRegister, data2, dat->dwCount);

  GL_VP_set_constant(priv, dat->dwRegister, dat->dwCount, (float *)data2);
  
  return D3D_OK;
}

HRESULT D3D_GL_delete_vertex_shader(GL_D3D_priv* priv, 
				    LPD3DHAL_DP2COMMAND cmd, 
				    LPVOID data1, 
				    LPVOID data2)
{
  D3DHAL_DP2VERTEXSHADER *dat = (D3DHAL_DP2VERTEXSHADER *)data1;
  VertexShader *vs = VS_FROM_HANDLE(dat->dwHandle);
  TRACE("priv(%p)->dwHandle(0x%lx)\n", priv, dat->dwHandle);
  
  if( IS_VS(dat->dwHandle) ) {
    if( priv->current_vs == vs ) {
      priv->current_vs = NULL;
    }
    
    if( vs->pFunction ) {
      GL_VP_delete(priv, (GLuint)vs->DriverReserved0);
    }
  }
  return D3D_OK;
}

#ifdef DEBUG_VERTEX_DATA
void dump_vertex(GL_D3D_priv* priv, DWORD dwIndex)
{
  VertexShader *vs = priv->current_vs;
  int i;
  D3DVALUE *ptr;
  DWORD *dptr;
  WORD *sptr;
  
  for( i=0; i<MAX_VERTEX_ATTRIBUTES; i++ ) {
    if( IS_INPUT_REG_ENABLED(vs->input_reg[i].type) ) {
      BYTE streamnum = vs->input_reg[i].stream;
      GLuint stride = priv->d3dp->stream[streamnum].stride;
      ptr = ((LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
	               + vs->input_reg[i].offset) + dwIndex * stride;
      sptr = (WORD *)ptr;
      dptr = (DWORD *)ptr;

      switch (GET_INPUT_REG_TYPE(vs->input_reg[i].type)) {
      case D3DVSDT_FLOAT1:
	TRACE("  reg %02d: %f\n", i, ptr[0]); break;
      case D3DVSDT_FLOAT2:
	TRACE("  reg %02d: %f, %f\n", i, ptr[0], ptr[1]); break;
      case D3DVSDT_FLOAT3:
	TRACE("  reg %02d: %f, %f, %f\n", i, ptr[0], ptr[1], ptr[2]); break;
      case D3DVSDT_FLOAT4:
	TRACE("  reg %02d: %f, %f, %f, %f\n", i, ptr[0], ptr[1], ptr[2], ptr[3]); break;
      case D3DVSDT_D3DCOLOR:
	TRACE("  reg %02d: %08lx\n", i, dptr[0]); break;
      case D3DVSDT_UBYTE4:
	TRACE("  reg %02d: %08lx\n", i, dptr[0]); break;
      case D3DVSDT_SHORT2:
	TRACE("  reg %02d: %d, %d\n", i, sptr[0], sptr[1]); break;
      case D3DVSDT_SHORT4:
	TRACE("  reg %02d: %d, %d, %d, %d\n", i, sptr[0], sptr[1], sptr[2], sptr[3]); break;
      }
    }
  }
}
#endif

void D3D_GL_setup_programmable(GL_D3D_priv* priv, LPVOID data1,
                               const WORD* indices, GLenum pt, DWORD dwCount, BOOL indexed)
{
  VertexShader *vs = priv->current_vs;
  GLuint i;
  DWORD dwIndexBase; 
  DWORD dwVStart, dwVCount;

  dwIndexBase = indexed ? ((LPD3DHAL_DP2STARTVERTEX)data1)->wVStart : 0;
  dwVStart = priv->dp2_dwVertexOffset;
  dwVCount = priv->dp2_dwVertexLength;

  TRACE("setup programmable (%p)\n", vs);

  /* disable the conventional arrays */
  if( priv->pVertexArray ) {
    glDisableClientState(GL_VERTEX_ARRAY);
    priv->pVertexArray = NULL;
    priv->sVertexArray = 0;
  }
  if( priv->pNormalArray ) {
    glDisableClientState(GL_NORMAL_ARRAY);
    priv->pNormalArray = NULL;
    priv->sNormalArray = 0;
  }
  if( priv->pColorArray ) {
    glDisableClientState(GL_COLOR_ARRAY);
    priv->pColorArray = NULL;
    priv->sColorArray = 0;
  }
  if( HAVE_GL_EXT(priv, EXT_secondary_color) && priv->pSecondaryColorArray ) {
    glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
    priv->pSecondaryColorArray = NULL;
    priv->sSecondaryColorArray = 0;
  }
  for( i=0; i<priv->lpCaps->wMaxTextureBlendStages; i++ ) {
    if( priv->pTexCoordArray[i] ) {
      GLD(priv)->glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      priv->pTexCoordArray[i] = NULL;
      priv->sTexCoordArray[i] = 0;
    }
  }
  if( HAVE_GL_EXT(priv, ARB_vertex_blend) ) {
    glDisableClientState(GL_WEIGHT_ARRAY_ARB);
  } else if( HAVE_GL_EXT(priv, EXT_vertex_weighting) ) {
    glDisableClientState(GL_VERTEX_WEIGHT_ARRAY_EXT);
  }

  /* setup the vertex attribute arrays */
  if( HAVE_GL_EXT(priv, NV_vertex_program) ) {
    
    for( i=0; i<MAX_VERTEX_ATTRIBUTES; i++ ) {
      GLenum GL_VERTEX_ATTRIB_ARRAYi_NV = GL_VERTEX_ATTRIB_ARRAY0_NV + i;
      
      if( IS_INPUT_REG_ENABLED(vs->input_reg[i].type) ) {
	BYTE streamnum = vs->input_reg[i].stream;
	GLsizei size = 0;
	GLenum type = GL_FLOAT;
	GLuint stride = priv->d3dp->stream[streamnum].stride;
	const GLvoid *pointer = ((LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
	               + vs->input_reg[i].offset) + (dwIndexBase * stride);
	TRACE("input reg%d: %d, 0x%02x, %d, %d, %ld, %p\n", i, streamnum,
	      GET_INPUT_REG_TYPE(vs->input_reg[i].type), 
	      vs->input_reg[i].offset, stride, dwIndexBase,
	      (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem);

	switch (GET_INPUT_REG_TYPE(vs->input_reg[i].type)) {
	case D3DVSDT_FLOAT1:
	  size = 1; type = GL_FLOAT; break;
	case D3DVSDT_FLOAT2:
	  size = 2; type = GL_FLOAT; break;
	case D3DVSDT_FLOAT3:
	  size = 3; type = GL_FLOAT; break;
	case D3DVSDT_FLOAT4:
	  size = 4; type = GL_FLOAT; break;
	case D3DVSDT_D3DCOLOR:
	  size = 4; type = GL_UNSIGNED_BYTE; break;
	case D3DVSDT_UBYTE4:
	  size = 4; type = GL_UNSIGNED_BYTE; break;
	case D3DVSDT_SHORT2:
	  size = 2; type = GL_SHORT; break;
	case D3DVSDT_SHORT4:
	  size = 4; type = GL_SHORT; break;
	default: 
	  FIXME("unknown input register data type %d\n", 
		GET_INPUT_REG_TYPE(vs->input_reg[i].type));
	}

	TRACE("enabling and setting vertex attrib pointer (%d,%d,%d,%d,%p)\n",
	      i, size, type, stride, pointer);
	
	glEnableClientState(GL_VERTEX_ATTRIB_ARRAYi_NV);
	GLD(priv)->glVertexAttribPointerNV(i, size, type, stride, pointer);
      } else {
	glDisableClientState(GL_VERTEX_ATTRIB_ARRAYi_NV);
      }
    }
  }

  TRACE("ibase=%ld, vstart=%ld, vcount=%ld\n", dwIndexBase, dwVStart, dwVCount);
  if( pt == GL_POINTS ) {
    /* FIXME: debug dumps for points */
  } else if( indexed ) {
    LPD3DHAL_DP2STARTVERTEX dat = (LPD3DHAL_DP2STARTVERTEX)data1;
    TRACE("start=%d, count=%ld [indexed]\n", dat->wVStart, dwCount);
#ifdef DEBUG_VERTEX_DATA
    {
      DWORD i;
      for( i=0; i<dwCount; i++ ) {
	WORD idx = indices[i];
	TRACE(" index%03ld: %03d\n", i, idx);
	dump_vertex(priv, dwIndexBase + idx);
      }
    }
#endif	  
  } else {
    LPD3DHAL_DP2STARTVERTEX dat = (LPD3DHAL_DP2STARTVERTEX)data1;
    TRACE("start = %d, count=%ld\n", dat->wVStart, dwCount);
#ifdef DEBUG_VERTEX_DATA
    {
      DWORD i;
      for( i=0; i<dwCount; i++) {
	dump_vertex(priv, dwVStart + dat->wVStart + i);
      }
    }
#endif
  }

  CHECK_GL_ERROR("setup programmable");
}

void D3D_GL_finish_programmable(GL_D3D_priv* priv) 
{
  GLuint i;

  if( HAVE_GL_EXT(priv, NV_vertex_program) ) {
    TRACE("disabling all vertex attribute arrays\n");
    for( i=0; i<MAX_VERTEX_ATTRIBUTES; i++ ) {
      glDisableClientState(GL_VERTEX_ATTRIB_ARRAY0_NV + i);
    }
  }
  CHECK_GL_ERROR("finish programmable");
}


