/*	Direct3DDevice interface
 *
 * Copyright 2001 TransGaming Technologies Inc.
 */

/* This file implements IDirect3DDevice, the first Direct3D interface
 * MS engineered, and the only interface that supports execute buffers.
 * Games written for IDirect3DDevice2/3 are allowed to QueryInterface
 * for IDirect3DDevice to use execute buffers, and many still do. */

#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "debugtools.h"
#include "ddrawi.h"
#include "d3dhal.h"
#include "d3d_private.h"
#include "d3ddevice/main.h"
#include "d3ddevice/exebuf.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define CONVERT7(pd3dd) COM_INTERFACE_CAST(IDirect3DDeviceImpl,	\
					   IDirect3DDevice,	\
					   IDirect3DDevice7,	\
					   (pd3dd))
#define CONVERT3(pd3dd) COM_INTERFACE_CAST(IDirect3DDeviceImpl,	\
					   IDirect3DDevice,	\
					   IDirect3DDevice3,	\
					   (pd3dd))
#define CONVERT2(pd3dd) COM_INTERFACE_CAST(IDirect3DDeviceImpl,	\
					   IDirect3DDevice,	\
					   IDirect3DDevice2,	\
					   (pd3dd))

static HRESULT WINAPI
Direct3DDevice1_QueryInterface(LPDIRECT3DDEVICE iface, REFIID riid,
				    LPVOID *ppObj)
{
    return IDirect3DDevice7_QueryInterface(CONVERT7(iface), riid, ppObj);
}

static ULONG WINAPI
Direct3DDevice1_AddRef(LPDIRECT3DDEVICE iface)
{
    return IDirect3DDevice7_AddRef(CONVERT7(iface));
}

static ULONG WINAPI
Direct3DDevice1_Release(LPDIRECT3DDEVICE iface)
{
    return IDirect3DDevice7_Release(CONVERT7(iface));
}

static HRESULT WINAPI
Direct3DDevice1_Initialize(LPDIRECT3DDEVICE iface, LPDIRECT3D lpDirect3D,
				LPGUID lpGUID, LPD3DDEVICEDESC lpD3DDVDesc)
{
    TRACE("(%p)->(%p,%p,%p)\n", iface, lpDirect3D, lpGUID, lpD3DDVDesc);
    return DDERR_ALREADYINITIALIZED;
}

static HRESULT WINAPI
Direct3DDevice1_GetCaps(LPDIRECT3DDEVICE iface,
			     LPD3DDEVICEDESC lpHAL,
			     LPD3DDEVICEDESC lpHEL)
{
    return IDirect3DDevice3_GetCaps(CONVERT3(iface), lpHAL, lpHEL);
}

static HRESULT WINAPI
Direct3DDevice1_AddViewport(LPDIRECT3DDEVICE iface,
				 LPDIRECT3DVIEWPORT lpView)
{
    /* IDirect3DViewportImpl has only one vtable */
    return IDirect3DDevice3_AddViewport(CONVERT3(iface),
					(LPDIRECT3DVIEWPORT3)lpView);
}

static HRESULT WINAPI
Direct3DDevice1_EnumTextureFormats(LPDIRECT3DDEVICE iface,
					LPD3DENUMTEXTUREFORMATSCALLBACK lpCB,
					LPVOID lpContext)
{
    return IDirect3DDevice2_EnumTextureFormats(CONVERT2(iface), lpCB, lpContext);
}

static HRESULT WINAPI
Direct3DDevice1_BeginScene(LPDIRECT3DDEVICE iface)
{
    return IDirect3DDevice7_BeginScene(CONVERT7(iface));
}

static HRESULT WINAPI
Direct3DDevice1_EndScene(LPDIRECT3DDEVICE iface)
{
    return IDirect3DDevice7_EndScene(CONVERT7(iface));
}

ICOM_VTABLE(IDirect3DDevice) Main_IDirect3DDevice1_VTable =
{
    Direct3DDevice1_QueryInterface,
    Direct3DDevice1_AddRef,
    Direct3DDevice1_Release,
    Direct3DDevice1_Initialize,
    Direct3DDevice1_GetCaps,
    0xdead7001, /* SwapTextureHandles */
    0xdead7002, /* CreateExecuteBuffer */
    0xdead7003, /* GetStats */
    0xdead7004, /* Execute */
    Direct3DDevice1_AddViewport,
    0xdead7005, /* DeleteViewport */
    0xdead7006, /* NextViewport */
    0xdead7007, /* Pick */
    0xdead7008, /* GetPickRecords */
    Direct3DDevice1_EnumTextureFormats,
    0xdead7009, /* CreateMatrix */
    0xdead700A, /* SetMatrix */
    0xdead700B, /* GetMatrix */
    0xdead700C, /* DeleteMatrix */
    Direct3DDevice1_BeginScene,
    Direct3DDevice1_EndScene,
    0xdead700D /* GetDirect3D */
};
