/*	Legacy Direct3DViewport interface
 *
 * Copyright 2001 TransGaming Technologies Inc.
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "debugtools.h"
#include "ddrawi.h"
#include "d3dhal.h"
#include "d3d_private.h"
#include "d3ddevice/main.h"
#include "d3ddevice/viewport.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

static ICOM_VTABLE(IDirect3DViewport3) IDirect3DViewport3_VTable;

/* Non-interface functions */
HRESULT
Direct3DViewport3_Construct(IDirect3DViewportImpl* This,
			    IDirectDrawImpl* pDD)
{
    This->ref = 1;
    This->d3d_owner = pDD;
    This->bd = NULL;

    ICOM_INIT_INTERFACE(This, IDirect3DViewport3,
			IDirect3DViewport3_VTable);

    /* Direct3D_AddViewport(This->d3d_owner, This); */

    return D3D_OK;
}

HRESULT
Direct3DViewport3_Create(IDirectDrawImpl *pDD,
			 LPDIRECT3DVIEWPORT3 *ppd3dv)
{
    IDirect3DViewportImpl* This;
    HRESULT hr;

    This = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
		     sizeof(*This));
    if (This == NULL) return E_OUTOFMEMORY;

    hr = Direct3DViewport3_Construct(This, pDD);
    if (FAILED(hr))
	HeapFree(GetProcessHeap(), 0, This);
    else
	*ppd3dv = ICOM_INTERFACE(This, IDirect3DViewport3);

    return hr;
}

static void Direct3DViewport3_Destroy(IDirect3DViewportImpl* This)
{
    /* Direct3D_RemoveViewport(This->d3d_owner, This); */
    HeapFree(GetProcessHeap(), 0, This);
}

void Direct3DViewport3_ForceDestroy(IDirect3DViewportImpl* This)
{
    WARN("destroying 3D material %p with refcnt %lu\n", This, This->ref);
    Direct3DViewport3_Destroy(This);
}

/* public non-interface functions */

void Direct3DViewport_Unbind(IDirect3DViewportImpl* This,
			     IDirect3DDeviceImpl* device)
{
    LPDIRECT3DDEVICE7 dev = ICOM_INTERFACE(device, IDirect3DDevice7);

    IDirect3DLightImpl* light;

    This->bound = FALSE;

    for (light = This->lights; light != NULL; light = light->next_view)
    {
	IDirect3DDevice7_LightEnable(dev, light->view_index, FALSE);
    }
}

void Direct3DViewport_Bind(IDirect3DViewportImpl* This,
			   IDirect3DDeviceImpl* device)
{
    LPDIRECT3DDEVICE7 dev = ICOM_INTERFACE(device, IDirect3DDevice7);

    IDirect3DLightImpl* light;

    This->bound = TRUE;

    IDirect3DDevice7_SetViewport(dev, &This->v);

    for (light = This->lights; light != NULL; light = light->next_view)
    {
	IDirect3DDevice7_SetLight(dev, light->view_index, &light->l);
	IDirect3DDevice7_LightEnable(dev, light->view_index, TRUE);
    }
}

/* Interface functions */
ULONG WINAPI Direct3DViewport3_Release(LPDIRECT3DVIEWPORT3 iface)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);

    if (--This->ref == 0)
    {
	Direct3DViewport3_Destroy(This);
	return 0;
    }

    return This->ref;
}

ULONG WINAPI Direct3DViewport3_AddRef(LPDIRECT3DVIEWPORT3 iface)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);

    return ++This->ref;
}

HRESULT WINAPI
Direct3DViewport3_QueryInterface(LPDIRECT3DVIEWPORT3 iface, REFIID riid,
				LPVOID* ppObj)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%s,%p)\n", This, debugstr_guid(riid), ppObj);

    if (IsEqualGUID(&IID_IUnknown, riid)
	|| IsEqualGUID(&IID_IDirect3DViewport3, riid)
	|| IsEqualGUID(&IID_IDirect3DViewport2, riid)
	|| IsEqualGUID(&IID_IDirect3DViewport, riid))
    {
	This->ref++;
	*ppObj = ICOM_INTERFACE(This, IDirect3DViewport3);
	return S_OK;
    }
    else
	return E_NOINTERFACE;
}

HRESULT WINAPI
Direct3DViewport3_Initialize(LPDIRECT3DVIEWPORT3 iface, LPDIRECT3D pd3d)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, pd3d);
    return DDERR_ALREADYINITIALIZED;
}

HRESULT WINAPI
Direct3DViewport3_GetViewport(LPDIRECT3DVIEWPORT3 iface, LPD3DVIEWPORT lpView)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, lpView);
    lpView->dwX      = This->v.dwX;
    lpView->dwY      = This->v.dwY;
    lpView->dwWidth  = This->v.dwWidth;
    lpView->dwHeight = This->v.dwHeight;
    lpView->dvScaleX = 1.0; /* ? */
    lpView->dvScaleY = 1.0; /* ? */
    lpView->dvMaxX   = 1.0; /* ? */
    lpView->dvMaxY   = 1.0; /* ? */
    lpView->dvMinZ   = This->v.dvMinZ;
    lpView->dvMaxZ   = This->v.dvMaxZ;
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_SetViewport(LPDIRECT3DVIEWPORT3 iface, LPD3DVIEWPORT lpView)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, lpView);
    This->v.dwX      = lpView->dwX;
    This->v.dwY      = lpView->dwY;
    This->v.dwWidth  = lpView->dwWidth;
    This->v.dwHeight = lpView->dwHeight;
    This->v.dvMinZ   = lpView->dvMinZ;
    This->v.dvMaxZ   = lpView->dvMaxZ;
    /* FIXME: dvScaleX/Y, dvMaxX/Y? */

    /* Hack. Unfortunately this is visible in GetViewport.
     * I'll have to check that against windows. */
    if (This->v.dvMinZ == 0 && This->v.dvMaxZ == 0)
	This->v.dvMaxZ = 1;

    if (This->dev_owner && This->dev_owner->view == This)
	IDirect3DDevice7_SetViewport(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
				     &This->v);

    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_TransformVertices(LPDIRECT3DVIEWPORT3 iface,
				    DWORD dwVertexCount,
				    LPD3DTRANSFORMDATA lpData,
				    DWORD dwFlags,
				    LPDWORD lpOffScreen)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    FIXME("(%p)->(%ld,%p,%lx,%p): stub\n", This, dwVertexCount, lpData, dwFlags, lpOffScreen);
    *lpOffScreen = 0;
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_LightElements(LPDIRECT3DVIEWPORT3 iface,
				DWORD dwElementCount,
				LPD3DLIGHTDATA lpData)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    /* this function was never implemented by MS */
    ERR("(%p)->(%ld,%p): unimplemented\n", This, dwElementCount, lpData);
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_SetBackground(LPDIRECT3DVIEWPORT3 iface,
				D3DMATERIALHANDLE hMat)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, (IDirect3DMaterialImpl*)hMat);
    This->bm = (IDirect3DMaterialImpl*)hMat;
    return DD_OK;
}


HRESULT WINAPI
Direct3DViewport3_GetBackground(LPDIRECT3DVIEWPORT3 iface,
				LPD3DMATERIALHANDLE lphMat, LPBOOL lpValid)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p,%p)\n", This, lphMat, lpValid);

    *lphMat = (D3DMATERIALHANDLE)This->bm;
    *lpValid = FALSE; /* ? */
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_SetBackgroundDepth(LPDIRECT3DVIEWPORT3 iface,
				LPDIRECTDRAWSURFACE lpDDSurface)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, (IDirectDrawSurfaceImpl*)lpDDSurface);
    This->bd = (IDirectDrawSurfaceImpl*)lpDDSurface;
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_Clear(LPDIRECT3DVIEWPORT3 iface, DWORD dwCount, LPD3DRECT lpRects,
			DWORD dwFlags)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    DWORD dwColor = 0;
    TRACE("(%p)->(%ld,%p,%08lx)\n", This, dwCount, lpRects, dwFlags);
    if ((dwFlags & D3DCLEAR_TARGET) && This->bm) {
	dwColor = D3DRGBA(This->bm->m.u.diffuse.u1.r,
			  This->bm->m.u.diffuse.u2.g,
			  This->bm->m.u.diffuse.u3.b,
			  This->bm->m.u.diffuse.u4.a);
    }
    if (This->dev_owner && This->dev_owner->view == This)
    {
		IDirect3DDevice7_Clear(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
			       dwCount,lpRects,dwFlags,dwColor,This->v.dvMaxZ,0);

		/* Blit the background depth if it exists */
		if (This->bd && This->dev_owner->zbuffer && (dwFlags & D3DCLEAR_ZBUFFER ))
		{
			IDirectDrawSurface3_Blt(ICOM_INTERFACE(This->dev_owner->zbuffer, IDirectDrawSurface3),
									NULL,
									ICOM_INTERFACE(This->bd, IDirectDrawSurface3),
									NULL, 0, 0);
		}
	}

    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_AddLight(LPDIRECT3DVIEWPORT3 iface, LPDIRECT3DLIGHT lpLight)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    IDirect3DLightImpl* light = ICOM_OBJECT(IDirect3DLightImpl,
					    IDirect3DLight,
					    lpLight);
    TRACE("(%p)->(%p)\n", This, lpLight);
    if (light->view_owner != NULL) return D3DERR_LIGHTHASVIEWPORT;
    /* allocate light - is the previous allocation still free? */
    if (This->lmask & (1 << light->view_index)) {
	/* it's taken, allocate another
	 * (hmm, we could use a better algorithm than this)
	 */
	unsigned u;
	for (u=0; u<32 && (This->lmask & (1 << u)); u++);
	if (u >= 32) {
	    ERR("too many allocated lights, can't attach!\n");
	    return D3D_OK;
	}
        light->view_index = u;
    }
    This->lmask |= 1 << light->view_index;

    /* attach to linked list */
    light->prev_view = NULL;
    if (This->lights)
      This->lights->prev_view = light;
    light->next_view = This->lights;
    This->lights = light;

    light->view_owner = This;
    if (This->dev_owner && This->dev_owner->view == This) {
	IDirect3DDevice7_SetLight(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
				  light->view_index, &light->l);
	IDirect3DDevice7_LightEnable(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
				     light->view_index, TRUE);
    }
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_DeleteLight(LPDIRECT3DVIEWPORT3 iface, LPDIRECT3DLIGHT lpLight)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    IDirect3DLightImpl* light = ICOM_OBJECT(IDirect3DLightImpl,
					    IDirect3DLight,
					    lpLight);
    TRACE("(%p)->(%p)\n", This, lpLight);
    if (light->view_owner != This) return D3DERR_LIGHTNOTINTHISVIEWPORT;
    if (This->dev_owner && This->dev_owner->view == This) {
	IDirect3DDevice7_LightEnable(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
				     light->view_index, FALSE);
    }
    This->lmask &= ~(1 << light->view_index);

    /* detach from linked list */
    if (light->prev_view)
      light->prev_view->next_view = light->next_view;
    else
      This->lights = light->next_view;
    if (light->next_view)
      light->next_view->prev_view = light->prev_view;

    light->view_owner = NULL;
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_NextLight(LPDIRECT3DVIEWPORT3 iface, LPDIRECT3DLIGHT lpDirect3DLight, LPDIRECT3DLIGHT *lplpDirect3DLight, DWORD dwFlags)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    IDirect3DLightImpl* light = ICOM_OBJECT(IDirect3DLightImpl,
					    IDirect3DLight,
					    lpDirect3DLight);
    TRACE("(%p)->(%p,%p,%08lx)\n", This, lpDirect3DLight, lplpDirect3DLight, dwFlags);
    if (light->view_owner != This) return D3DERR_LIGHTNOTINTHISVIEWPORT;

    *lplpDirect3DLight = ICOM_INTERFACE(light->next_view, IDirect3DLight);
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_GetViewport2(LPDIRECT3DVIEWPORT3 iface, LPD3DVIEWPORT2 lpView)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, lpView);
    lpView->dwX          = This->v.dwX;
    lpView->dwY          = This->v.dwY;
    lpView->dwWidth      = This->v.dwWidth;
    lpView->dwHeight     = This->v.dwHeight;
    lpView->dvClipX      = 0.0; /* ? */
    lpView->dvClipY      = 0.0; /* ? */
    lpView->dvClipWidth  = 1.0; /* ? */
    lpView->dvClipHeight = 1.0; /* ? */
    lpView->dvMinZ       = This->v.dvMinZ;
    lpView->dvMaxZ       = This->v.dvMaxZ;
    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_SetViewport2(LPDIRECT3DVIEWPORT3 iface, LPD3DVIEWPORT2 lpView)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%p)\n", This, lpView);
    This->v.dwX      = lpView->dwX;
    This->v.dwY      = lpView->dwY;
    This->v.dwWidth  = lpView->dwWidth;
    This->v.dwHeight = lpView->dwHeight;
    This->v.dvMinZ   = lpView->dvMinZ;
    This->v.dvMaxZ   = lpView->dvMaxZ;
    /* FIXME: dvClipX/Y/W/H? */

    /* Hack. Unfortunately this is visible in GetViewport.
     * I'll have to check that against windows. */
    if (This->v.dvMinZ == 0 && This->v.dvMaxZ == 0)
	This->v.dvMaxZ = 1;

    if (This->dev_owner && This->dev_owner->view == This)
	IDirect3DDevice7_SetViewport(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
				     &This->v);

    return DD_OK;
}

HRESULT WINAPI
Direct3DViewport3_Clear2(LPDIRECT3DVIEWPORT3 iface, DWORD dwCount, LPD3DRECT lpRects,
			 DWORD dwFlags, DWORD dwColor, D3DVALUE dvZ, DWORD dwStencil)
{
    ICOM_THIS(IDirect3DViewportImpl, iface);
    TRACE("(%p)->(%ld,%p,%08lx)\n", This, dwCount, lpRects, dwFlags);
    if (This->dev_owner && This->dev_owner->view == This)
	IDirect3DDevice7_Clear(ICOM_INTERFACE(This->dev_owner, IDirect3DDevice7),
			       dwCount,lpRects,dwFlags,dwColor,dvZ,dwStencil);
    return DD_OK;
}

static ICOM_VTABLE(IDirect3DViewport3) IDirect3DViewport3_VTable =
{
    Direct3DViewport3_QueryInterface,
    Direct3DViewport3_AddRef,
    Direct3DViewport3_Release,
    Direct3DViewport3_Initialize,
    Direct3DViewport3_GetViewport,
    Direct3DViewport3_SetViewport,
    Direct3DViewport3_TransformVertices,
    Direct3DViewport3_LightElements,
    Direct3DViewport3_SetBackground,
    Direct3DViewport3_GetBackground,
    Direct3DViewport3_SetBackgroundDepth,
    0xdead3008, /* GetBackgroundDepth */
    Direct3DViewport3_Clear,
    Direct3DViewport3_AddLight,
    Direct3DViewport3_DeleteLight,
    Direct3DViewport3_NextLight,
    Direct3DViewport3_GetViewport2,
    Direct3DViewport3_SetViewport2,
    0xdead300F, /* SetBackgroundDepth2 */
    0xdead3010, /* GetBackgroundDepth2 */
    Direct3DViewport3_Clear2
};
