/* dpnet.dll
 *
 * Copyright 2002, 2003 TransGaming Technologies
 *
 * David Hammerton (david@transgaming.com)
 */

#include "windef.h"
#include "winbase.h"
#include "winreg.h"
#include "objbase.h"
#include "debugtools.h"
#include "winerror.h"
#include "winnt.h"
#include "wine/unicode.h"

#include "dplay8.h"
#include "dpaddr.h"
#include "dplay8_private.h"

DEFAULT_DEBUG_CHANNEL(dplay);

extern IDirectPlay8ServiceProvider *globalSP; /* hack. see general.c to as why */

static ICOM_VTABLE(IDirectPlay8Address) directPlay8AddressVT;
static ICOM_VTABLE(IDirectPlay8AddressIP) directPlay8AddressIPVT;

HRESULT WINAPI DirectPlay8Address_QueryInterface(PDIRECTPLAY8ADDRESS iface,
                                                 REFIID riid, LPVOID *obj)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);
    TRACE("(%p)->(%s, %p)\n", iface, debugstr_guid(riid), obj);
    if (IsEqualGUID(&IID_IUnknown, riid) ||
        IsEqualGUID(&IID_IDirectPlay8Address, riid))
    {
        This->ref++;
        *obj = This;
        return S_OK;
    }
    else if (IsEqualGUID(&IID_IDirectPlay8AddressIP, riid))
    {
        This->ref++;
        *obj = (LPVOID *)This->lpVtbl2;
        return S_OK;
    }
    else
    {
        FIXME("(%p)->(%s, %p): no interface\n", iface, debugstr_guid(riid), obj);
        return E_NOINTERFACE;
    }
}

ULONG WINAPI DirectPlay8Address_AddRef(PDIRECTPLAY8ADDRESS iface)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);
    TRACE("(%p)->()\n", iface);
    return ++This->ref;
}

ULONG WINAPI DirectPlay8Address_Release(PDIRECTPLAY8ADDRESS iface)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);
    TRACE("(%p)->()\n", iface);
    if (--This->ref) return This->ref;

    if (This->spData && !This->spDataRelease) FIXME("no release for data @ %p\n", This->spData);
    if (This->spData) This->spDataRelease(This->spData);
    HeapFree(GetProcessHeap(), 0, This->url);
    HeapFree(GetProcessHeap(), 0, This);
    return 0;
}

HRESULT WINAPI DirectPlay8Address_BuildFromURLA(PDIRECTPLAY8ADDRESS iface, CHAR *pszSourceURL)
{
    /* ICOM_THIS(IDirectPlay8AddressImpl, iface); */

    WCHAR *pwszSourceURL;
    INT iUrlStringLen;
    HRESULT hr;

    TRACE("(%p)->(%s)\n", iface, debugstr_a(pszSourceURL));

    iUrlStringLen = strlen(pszSourceURL);
    pwszSourceURL = HeapAlloc(GetProcessHeap(), 0, (iUrlStringLen + 1) * sizeof(WCHAR));
    MultiByteToWideChar(CP_ACP, 0, pszSourceURL, iUrlStringLen, pwszSourceURL,
                        (iUrlStringLen + 1) * sizeof(WCHAR) / sizeof(CHAR));

    hr = iface->lpVtbl->BuildFromURLW(iface, pwszSourceURL);

    HeapFree(GetProcessHeap(), 0, pwszSourceURL);

    return hr;
}

HRESULT WINAPI DirectPlay8Address_BuildFromURLW(PDIRECTPLAY8ADDRESS iface, WCHAR *pwszSourceURL)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);
    WCHAR *provider = NULL;
    DWORD providerLen = 0;
    GUID guidSP;
    HRESULT hr;

    TRACE("(%p)->(%s)\n", iface, debugstr_w(pwszSourceURL));

    if (strncmpW(pwszSourceURL, DPNA_HEADER, strlenW(DPNA_HEADER)) != 0)
        return DPNERR_INVALIDURL;

    if (This->url) HeapFree(GetProcessHeap(), 0, This->url);
    This->url = DPNET_strdupW(pwszSourceURL);

    if (DPNET_GetKeyValueW(This->url, DPNA_KEY_PROVIDER, provider, &providerLen)
            != DPNERR_BUFFERTOOSMALL)
    {
        HeapFree(GetProcessHeap(), 0, This->url);
        This->url = NULL;
        return DPNERR_INVALIDURL;
    }

    provider = HeapAlloc(GetProcessHeap(), 0, providerLen * sizeof(WCHAR));
    DPNET_GetKeyValueW(This->url, DPNA_KEY_PROVIDER, provider, &providerLen);
    CLSIDFromString(provider, &guidSP);
    HeapFree(GetProcessHeap(), 0, provider);
    iface->lpVtbl->SetSP(iface, &guidSP);
    TRACE("guid of provider: %s\n", debugstr_guid(&guidSP));

    if (!globalSP)
    {
        ERR("eek! no service provider selected! fix this hack.\n");
        HeapFree(GetProcessHeap(), 0, This->url);
        This->url = NULL;
        return DPNERR_INVALIDURL;
    }
    hr = globalSP->lpVtbl->SetSPAddressData(globalSP, iface);

    if (FAILED(hr))
    {
        HeapFree(GetProcessHeap(), 0, This->url);
        This->url = NULL;
        return DPNERR_INVALIDURL;
    }

    return S_OK;
}

HRESULT WINAPI DirectPlay8Address_Duplicate(PDIRECTPLAY8ADDRESS iface, PDIRECTPLAY8ADDRESS *ppdpaNewAddress)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);
    HRESULT hr;
    IDirectPlay8AddressImpl *newAddressThis;

    TRACE("(%p)->(%p)\n", iface, ppdpaNewAddress);
    hr = CoCreateInstance(&CLSID_DirectPlay8Address, NULL, CLSCTX_INPROC_SERVER,
                          &IID_IDirectPlay8Address,
                          (LPVOID *) ppdpaNewAddress);
    if (!SUCCEEDED(hr))
    {
        TRACE("duplication failed\n");
        return DPNERR_GENERIC;
    }

    newAddressThis = (IDirectPlay8AddressImpl *)(*ppdpaNewAddress);

    memcpy(&newAddressThis->guidSP, &This->guidSP, sizeof(GUID));
    /* FIXME: SHOULD USE URLS */
    if (This->spData) This->spDataDuplicate(*ppdpaNewAddress, This->spData);
    TRACE("dupilcated\n");

    return S_OK;
}

HRESULT WINAPI DirectPlay8Address_GetURLW(PDIRECTPLAY8ADDRESS iface, WCHAR *pwszURL, PDWORD pdwNumChars)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);
    DWORD numCharsRequired = 0;

    TRACE("(%p)->(%p, %p (%li))\n", iface, pwszURL, pdwNumChars, *pdwNumChars);

    if (!This->url) return DPNERR_INVALIDURL;
    if (pdwNumChars == NULL) return DPNERR_GENERIC;
    numCharsRequired = strlenW(This->url);

    if (*pdwNumChars < numCharsRequired)
    {
        *pdwNumChars = numCharsRequired;
        return DPNERR_BUFFERTOOSMALL;
    }
    strcpyW(pwszURL, This->url);

    return S_OK;
}

HRESULT WINAPI DirectPlay8Address_SetSP(PDIRECTPLAY8ADDRESS iface, const GUID *const pguidSP)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);

    TRACE("(%p)->(%s)\n", iface, debugstr_guid(pguidSP));
    memcpy(&This->guidSP, pguidSP, sizeof(GUID));
    return S_OK;
}

HRESULT DPNET_CreateDirectPlay8Address(IUnknown *pOuter, REFIID riid, LPVOID *ppobj)
{
    IDirectPlay8AddressImpl *ipDP8A;
    HRESULT hr;
    TRACE("()\n");
    ipDP8A = (IDirectPlay8AddressImpl *) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
                                                   sizeof(IDirectPlay8AddressImpl));
    if (ipDP8A == NULL) return E_OUTOFMEMORY;
    ICOM_VTBL(ipDP8A) = &directPlay8AddressVT;
    ipDP8A->lpVtbl2 = &directPlay8AddressIPVT;
    IDirectPlay8Address_AddRef((IDirectPlay8Address *)ipDP8A);
    TRACE("Created new object: %p\n", ipDP8A);
    hr = IDirectPlay8Address_QueryInterface((IDirectPlay8Address *)ipDP8A, riid, ppobj);
    IDirectPlay8Address_Release((IDirectPlay8Address *)ipDP8A);

    ipDP8A->hDialogMutex = CreateMutexA(NULL, FALSE, "DPNET Address DialogMutex");
    if (SUCCEEDED(hr))
        return S_OK;
    else
        return E_NOINTERFACE;
}

/* DirectPlayAddressIP wrappers */
HRESULT WINAPI DirectPlay8AddressIP_QueryInterface(PDIRECTPLAY8ADDRESSIP iface,
                                                   REFIID riid, LPVOID *obj)
{
    ICOM_THIS_MULTI(IDirectPlay8AddressImpl, lpVtbl2, iface);
    TRACE("(%p)->(%s, %p)\n", iface, debugstr_guid(riid), obj);
    return IDirectPlay8Address_QueryInterface((PDIRECTPLAY8ADDRESS) This, riid, obj);
}

ULONG WINAPI DirectPlay8AddressIP_AddRef(PDIRECTPLAY8ADDRESSIP iface)
{
    ICOM_THIS_MULTI(IDirectPlay8AddressImpl, lpVtbl2, iface);
    TRACE("(%p)->()\n", iface);
    return IDirectPlay8Address_AddRef((PDIRECTPLAY8ADDRESS) This);
}

ULONG WINAPI DirectPlay8AddressIP_Release(PDIRECTPLAY8ADDRESSIP iface)
{
    ICOM_THIS_MULTI(IDirectPlay8AddressImpl, lpVtbl2, iface);
    TRACE("(%p)->()\n", iface);
    return IDirectPlay8Address_Release((PDIRECTPLAY8ADDRESS) This);
}


static ICOM_VTABLE(IDirectPlay8Address) directPlay8AddressVT =
{
    ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
    DirectPlay8Address_QueryInterface,
    DirectPlay8Address_AddRef,
    DirectPlay8Address_Release,
    DirectPlay8Address_BuildFromURLW,
    DirectPlay8Address_BuildFromURLA,
    DirectPlay8Address_Duplicate,
    0xdead7007, /* SetEqual */
    0xdead7008, /* IsEqual */
    0xdead7009, /* Clear */
    DirectPlay8Address_GetURLW,
    0xdead700b, /* GetURLA */
    0xdead700c, /* GetSP */
    0xdead700d, /* GetUserData */
    DirectPlay8Address_SetSP,
    0xdead700f, /* SetUserData */
    0xdead7010, /* GetNumComponents */
    0xdead7011, /* GetComponentByName */
    0xdead7012, /* GetComponentByIndex */
    0xdead7012, /* AddComponent */
    0xdead7013, /* SetDevice */
    0xdead7014, /* GetDevice */
    0xdead7015  /* BuildFromDirectPlay4Address */
};

static ICOM_VTABLE(IDirectPlay8AddressIP) directPlay8AddressIPVT =
{
    ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
    DirectPlay8AddressIP_QueryInterface,
    DirectPlay8AddressIP_AddRef,
    DirectPlay8AddressIP_Release,
    0xdead7024, /* BuildFromSockAddr */
    0xdead7025, /* BuildFromAddress */
    0xdead7026, /* BuildLocalAddress */
    0xdead7027, /* GetSockAddress */
    0xdead7028, /* GetLocalAddress */
    0xdead7029  /* GetAddress */
};

/* private functions. used by the rest of dpnet.
 * Ideally should be in an interface */
HRESULT DPNET_Address_SetSPData(PDIRECTPLAY8ADDRESS iface, void *data,
                                spReleaseAddressInfo spReleaseCallback,
                                spDuplicateAddressInfo spDuplicateCallback)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);

    This->spData = data;
    This->spDataRelease = spReleaseCallback;
    This->spDataDuplicate = spDuplicateCallback;
    return S_OK;
}

HRESULT DPNET_Address_GetSPData(PDIRECTPLAY8ADDRESS iface, void **data)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);

    *data = This->spData;
    return S_OK;
}

HRESULT DPNET_Address_GetDialogLock(PDIRECTPLAY8ADDRESS iface)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);

    if (WaitForSingleObject(This->hDialogMutex, INFINITE) != WAIT_OBJECT_0)
        return E_FAIL;
    return S_OK;
}

HRESULT DPNET_Address_ReleaseDialogLock(PDIRECTPLAY8ADDRESS iface)
{
    ICOM_THIS(IDirectPlay8AddressImpl, iface);

    ReleaseMutex(This->hDialogMutex);
    return S_OK;
}

/* utility functions for parsing URLs */

HRESULT DPNET_GetKeyValueW(const WCHAR *url, const WCHAR *keyName,
                           WCHAR *keyValue /* [OUT] */,
                           PDWORD bufferLen /* [IN OUT] */)
{
    WCHAR *p, *p2, *p3;
    DWORD requiredLen;

    TRACE("%s, %s, %s, %p (%li)\n", debugstr_w(url), debugstr_w(keyName), debugstr_w(keyValue),
          bufferLen, *bufferLen);

    p = strstrW(url, keyName);

    if (p == NULL) return DPNERR_DOESNOTEXIST;

    p = strchrW(p, DPNA_SEPARATOR_KEYVALUE);

    if (p == NULL) return DPNERR_DOESNOTEXIST;
    p++;

    p2 = strchrW(p, DPNA_SEPARATOR_COMPONENT);
    p3 = strchrW(p, DPNA_SEPARATOR_USERDATA);
    if (p2 == NULL && p3 == NULL) /* end of string */
        requiredLen = strlenW(p);
    else if (p2 == NULL)
        requiredLen = p3 - p;
    else if (p3 == NULL)
        requiredLen = p2 - p;
    else
        requiredLen = (p2 < p3 ? p2 : p3) - p;

    if (*bufferLen < requiredLen)
    {
        *bufferLen = requiredLen;
        return DPNERR_BUFFERTOOSMALL;
    }
    strncpyW(keyValue, p, requiredLen);
    keyValue[requiredLen] = 0;

    TRACE("returning %s (%li).\n", debugstr_w(keyValue), requiredLen);

    return DPN_OK;
}

HRESULT DPNET_GetKeyValueFromAddressW(IDirectPlay8Address *pAddr, CONST WCHAR *keyName,
                                      WCHAR *keyValue /* [OUT] */,
                                      PDWORD bufferLen /* [IN OUT] */)
{
    ICOM_THIS(IDirectPlay8AddressImpl, pAddr);

    TRACE("%p, %s, %s, %p (%li)\n", pAddr, debugstr_w(keyName), debugstr_w(keyValue),
          bufferLen, *bufferLen);

    return DPNET_GetKeyValueW(This->url, keyName, keyValue, bufferLen);
}


