/*
 * DirectDraw driver interface
 *
 * Copyright 2001 TransGaming Technologies, Inc.
 */

#include "config.h"

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/poll.h>

#include "ts_xlib.h"
#include "x11drv.h"
#include "x11ddraw.h"
#include "xvidmode.h"
#include "dga2.h"
#include "glx.h"
#include "wine_gl.h"
#include "osmesa.h"

#include "windef.h"
#include "wingdi.h"
#include "ddrawi.h"
#include "bitmap.h"
#include "debugtools.h"
#include "options.h"
#include "mwm.h"

#define OLD_CRAP
#define NEW_STUFF
#define DO_DIB
#define FULLSCREEN

DEFAULT_DEBUG_CHANNEL(x11drv);

extern XContext winContext;
extern Atom mwmHints;
extern Atom netWMState;
extern Atom netWMStateStaysOnTop;
extern Atom winWMLayer;

extern Bool EVENT_CheckEvent( Display *display, XEvent *event_return );
extern Bool EVENT_CheckIsEvent( Display *display, int type, Window window );

extern int dxgrab;
extern int X11DRV_NoCursor;

extern int app_resolution_change;

LPDDRAWI_DDRAWSURFACE_LCL X11DRV_DD_Primary;
LPDDRAWI_DDRAWSURFACE_GBL X11DRV_DD_PrimaryGbl;
HWND X11DRV_DD_DeviceWnd;
Window X11DRV_DD_DeviceWindow;
#ifdef OLD_CRAP
HWND X11DRV_DD_PrimaryWnd;
ATOM X11DRV_DD_UserClass;
#endif
HBITMAP X11DRV_DD_PrimaryDIB;
Drawable X11DRV_DD_PrimaryDrawable;
BOOL X11DRV_DD_IsDirect;
Window X11DRV_DD_PrimaryFrame, X11DRV_DD_PrimaryClient;
HANDLE X11DRV_DD_Thread, X11DRV_DD_SyncH;
int X11DRV_DD_Pipe[2];
DWORD X11DRV_DD_Busy, X11DRV_DD_Paint, X11DRV_DD_Lock, X11DRV_DD_RefCount;
DWORD X11DRV_DD_DeviceWndOK;
RECT X11DRV_DD_OldWindowRect, X11DRV_DD_OldClientRect;
Window X11DRV_DD_OldRootWindow;

#define DDMSG_QUIT    0
#define DDMSG_SYNC    1
#define DDMSG_UPDATE  2
#define DDMSG_GRAB    3
#define DDMSG_UNGRAB  4

#ifdef NEW_STUFF
/*
 * Window manager issues:
 *  KDE/KWIN: to get above panel, window must be decorationless and
 *            have the same geometry as the desktop
 *  or else:  the _NET_WM_STATE property must be set to _NET_WM_STATE_STAYS_ON_TOP
 *  GNOME:    the _WIN_LAYER property must be set to ABOVE_DOCK
 *
 * FIXME: create window using desktop thread, not app thread
 */
static void X11DRV_DD_CreateWindow(unsigned int width, unsigned int height)
{
  Display *display = thread_display();
  char *name = "Wine DirectDraw HAL";
  XSizeHints *size_hints;
  XWMHints   *wm_hints;
  XClassHint *class_hints;
  XSetWindowAttributes win_attr;
  XTextProperty window_name;
  Window frame, client;
  unsigned int fwidth, fheight;

  wine_tsx11_lock();

  /* Create (full-screen) frame window */
  win_attr.background_pixel = BlackPixel(display, 0);
  win_attr.event_mask = /* ExposureMask | */ KeyPressMask | KeyReleaseMask |
    PointerMotionMask | ButtonPressMask | ButtonReleaseMask |
    StructureNotifyMask | FocusChangeMask | KeymapStateMask;
  if (root_window == DefaultRootWindow(display))
    win_attr.cursor = X11DRV_GetCursor(display, X11DRV_NoCursor ? NULL : GlobalLock16(GetCursor()));
  else
    win_attr.cursor = None; /* inherit Wine desktop cursor */
  win_attr.colormap = X11DRV_PALETTE_PaletteXColormap;
  win_attr.override_redirect = !Options.managed;

#ifdef FULLSCREEN
  fwidth = default_screen_width;
  fheight = default_screen_height;
#else
  fwidth = width;
  fheight = height;
#endif

  frame = XCreateWindow( display, root_window,
			 0, 0, fwidth, fheight, 0, screen_depth, InputOutput, visual,
			 CWBackPixel | CWEventMask | CWCursor | CWColormap |
			 CWOverrideRedirect, &win_attr );

  /* Set window manager properties */
  size_hints  = XAllocSizeHints();
  wm_hints    = XAllocWMHints();
  class_hints = XAllocClassHint();
  if (!size_hints || !wm_hints || !class_hints) {
    MESSAGE("Not enough memory for window manager hints.\n" );
    ExitProcess(1);
  }
  size_hints->min_width = size_hints->max_width = fwidth;
  size_hints->min_height = size_hints->max_height = fheight;
  size_hints->flags = PMinSize | PMaxSize | USPosition | USSize;

  wm_hints->flags = InputHint | StateHint;
  wm_hints->input = True; /* accept focus */
  wm_hints->initial_state = NormalState;
  class_hints->res_name  = "wine";
  class_hints->res_class = "Wine";

  XStringListToTextProperty( &name, 1, &window_name );
  XSetWMProperties( display, frame, &window_name, &window_name,
		    NULL, 0, size_hints, wm_hints, class_hints );
  if (win_attr.cursor != None) XFreeCursor(display, win_attr.cursor);
  XFree( size_hints );
  XFree( wm_hints );
  XFree( class_hints );

  /* no decorations! */
  if (mwmHints != None) {
    MwmHints mwm_hints;
    mwm_hints.flags = MWM_HINTS_FUNCTIONS | MWM_HINTS_DECORATIONS;
    mwm_hints.functions = 0;
    mwm_hints.decorations = 0;
    XChangeProperty( display, frame, mwmHints, mwmHints, 32,
		     PropModeReplace, (char*)&mwm_hints, sizeof(mwm_hints)/sizeof(long) );
  }

  /* tell KDE-compliant WMs that we want to stay on top
   * (though this does not really have any effect on fullscreen windows) */
  if (netWMState != None) {
    Atom state = netWMStateStaysOnTop;
    XChangeProperty( display, frame, netWMState,
		     XA_ATOM, 32,
		     PropModeReplace, (char*)&state, 1);
  }

  /* tell GNOME-compliant WMs that we want to stay on top */
  if (winWMLayer != None) {
    long layer = 10; /* ABOVE_DOCK */
    XChangeProperty( display, frame, winWMLayer,
		     XA_CARDINAL, 32,
		     PropModeReplace, (char*)&layer, 1);
  }

  /* Create real ddraw window as a child of the frame window,
   * sized at the requested resolution; this is done so that
   * 1) the frame window gets above all desktop panels
   *    (in KDE 2 it must be fullscreen to do this)
   * 2) the child window gets the benefit of the frame window,
   *    but a mouse grab on it will limit mouse movement to
   *    the resolution the app really wanted
   * then we can use XF86VidMode to only display the child
   * window, and the resolution should then be perfectly faked,
   * and no GNOME/KDE desktop stuff in the way */
  win_attr.background_pixel = BlackPixel(display, 0);
  win_attr.event_mask = ExposureMask /* | KeyPressMask | KeyReleaseMask |
    PointerMotionMask | ButtonPressMask | ButtonReleaseMask */ ;
  win_attr.cursor = None; /* inherit frame cursor */
  win_attr.colormap = X11DRV_PALETTE_PaletteXColormap;

  client = XCreateWindow( display, frame,
			  0, 0, width, height, 0, screen_depth, InputOutput, visual,
			  CWBackPixel | CWEventMask | CWCursor | CWColormap, &win_attr );

  X11DRV_DD_PrimaryFrame = frame;
  X11DRV_DD_PrimaryClient = client;

  XMapWindow( display, client );
  XMapWindow( display, frame );

  XFlush( display );
  wine_tsx11_unlock();
}

static void X11DRV_DD_DestroyWindow(void)
{
  Display *display = thread_display();

  wine_tsx11_lock();
  XSync( gdi_display, False );
  XUnmapWindow( display, X11DRV_DD_PrimaryFrame );
  XUnmapWindow( display, X11DRV_DD_PrimaryClient );
  XDestroyWindow( display, X11DRV_DD_PrimaryClient );
  X11DRV_DD_PrimaryClient = 0;
  XDestroyWindow( display, X11DRV_DD_PrimaryFrame );
  X11DRV_DD_PrimaryFrame = 0;
  XFlush( display );
  wine_tsx11_unlock();
}

static void X11DRV_DD_UpdateWindow(BOOL flip_sync)
{
  Display *display = gdi_display;
  Window win = X11DRV_DD_PrimaryClient;
  LPDDRAWI_DDRAWSURFACE_GBL gbl = X11DRV_DD_PrimaryGbl;
  HBITMAP hBmp = GET_LPDDRAWSURFACE_GBL_MORE(gbl)->hKernelSurface;
  BITMAPOBJ *bmp;
  X11DRV_DIBSECTION *dib;

  int state;

  if (!hBmp) {
    ERR("no DIB in surface\n");
    return;
  }
  bmp = (BITMAPOBJ *)GDI_GetObjPtr(hBmp, BITMAP_MAGIC);
  dib = (X11DRV_DIBSECTION *)bmp->dib;
  wine_tsx11_lock();
  state = X11DRV_DIB_Lock(bmp, DIB_Status_None, FALSE);

  /* FIXME: delegate this to another thread, for speed */

  switch (state) {
  case DIB_Status_AppMod:
    X11DRV_DIB_DrawDIBSection(bmp, win, 0, 0, 0, 0,
			      gbl->wWidth, gbl->wHeight);
    break;
  case DIB_Status_InSync:
  case DIB_Status_GdiMod:
    XCopyArea(display, (Drawable)bmp->physBitmap,
	      win, BITMAP_GC(bmp), 0, 0,
	      gbl->wWidth, gbl->wHeight,
	      0, 0);
    XSync(display, False);
    break;
  case DIB_Status_AuxMod:
    /* if this ever happens, there's probably no updating to do,
     * the sub-driver should take care of the updates */
    break;
  }

  X11DRV_DIB_Unlock(bmp, TRUE);
  wine_tsx11_unlock();
  GDI_ReleaseObj(hBmp);
}

void X11DRV_DD_Update(void);

void X11DRV_DD_ThreadEvent(Display *display, XEvent *event)
{
  Window target = X11DRV_DD_DeviceWindow;

  switch (event->type) {
  case KeyPress:
    /* cut down on autorepeat (assumes detectable autorepeat);
     * games rarely care about it, and it slows us down */
    {
      XEvent nevent;
      if (EVENT_CheckEvent(display, &nevent) &&
          nevent.xany.type == KeyPress &&
          nevent.xkey.window == event->xkey.window &&
          nevent.xkey.keycode == event->xkey.keycode)
        break;
    }
    /* fall through */
  case KeyRelease:
#if 1
    event->xany.window = root_window;
#else
    event->xany.window = X11DRV_DD_DeviceWindow;
#endif
    wine_tsx11_lock();
    if (!target) {
      int rev;
      XGetInputFocus(display, &target, &rev);
    }
    XSendEvent(display, target, True, KeyPressMask | KeyReleaseMask, event);
    wine_tsx11_unlock();
    break;
  case MotionNotify:
    /* cut down on motion events, they slow us down */
    if (EVENT_CheckIsEvent(display, MotionNotify, event->xany.window))
      break;
    /* fall through */
  case ButtonPress:
  case ButtonRelease:
    event->xany.window = root_window;
    wine_tsx11_lock();
    if (!target) {
      int dx, dy;
      XTranslateCoordinates(display, X11DRV_DD_PrimaryFrame, X11DRV_DD_PrimaryClient,
                            event->xmotion.x, event->xmotion.y,
                            &dx, &dy, &target);
      event->xmotion.x = dx;
      event->xmotion.y = dy;
    }
    if (target) {
      TRACE("forwarding mouse event %d to %08lx, pos (%d,%d)\n",
            event->xany.type, target, event->xmotion.x, event->xmotion.y);
      XSendEvent(display, target, True, PointerMotionMask, event);
    } else {
      TRACE("ignoring mouse event %d, pos (%d,%d)\n",
            event->xany.type, event->xmotion.x, event->xmotion.y);
    }
    wine_tsx11_unlock();
    break;
  case KeymapNotify:
    event->xany.window = root_window;
    wine_tsx11_lock();
    if (!target) {
      int rev;
      XGetInputFocus(display, &target, &rev);
    }

    /* Forward this event, unless it was already forwarded by us */
    if( !event->xany.send_event )
    {
      TRACE("forwarding keyboard event %d to %08lx\n",
            event->xany.type, target);
      XSendEvent(display, target, True, KeymapStateMask, event);
    }
    wine_tsx11_unlock();
    break;
  case Expose:
    X11DRV_DD_Update();
    break;
  case FocusIn:
    TRACE("acquired focus: %08lx %d %d\n", event->xfocus.window, event->xfocus.mode, event->xfocus.detail);
    /* X11DRV_EnableGrabs(); */ /* in case the grab has been lost */
    break;
  case FocusOut:
    TRACE("lost focus: %08lx %d %d\n", event->xfocus.window, event->xfocus.mode, event->xfocus.detail);
    break;
  }
}

BOOL X11DRV_DD_ThreadCmd(Display *display, int cmd)
{
  switch (cmd) {
  case DDMSG_QUIT:
    return TRUE;
  case DDMSG_SYNC:
    SetEvent(X11DRV_DD_SyncH);
    break;
  case DDMSG_UPDATE:
    if (!X11DRV_DD_Paint) break;
    X11DRV_DD_Busy++;
    X11DRV_DD_Paint = 0;
    if (X11DRV_DD_PrimaryDIB)
      X11DRV_DD_UpdateWindow(FALSE);
    X11DRV_DD_Busy--;
    break;
  case DDMSG_GRAB:
    TSXGrabPointer(display, X11DRV_DD_PrimaryFrame, True,
		   PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
		   GrabModeAsync, GrabModeAsync,
		   X11DRV_DD_PrimaryClient, None, CurrentTime);
    break;
  case DDMSG_UNGRAB:
    TSXUngrabPointer(display, CurrentTime);
    break;
  default:
    ERR("unknown command\n");
    return TRUE;
  }
  return FALSE;
}

static DWORD CALLBACK X11DRV_DD_ThreadProc(LPVOID arg)
{
  int *input = arg;
  Display *display = thread_display();
  int conn = ConnectionNumber(display);
  unsigned int width, height;
  struct pollfd fds[2];

  read(*input, &width, sizeof(width));
  read(*input, &height, sizeof(height));
  X11DRV_DD_CreateWindow(width, height);
  SetEvent(X11DRV_DD_SyncH);
  fds[0].fd = conn;
  fds[0].events = POLLIN;
  fds[1].fd = *input;
  fds[1].events = POLLIN;
  do {
    fds[0].revents = 0;
    if (poll(fds, 2, -1) < 0) break;
    if (fds[0].revents & POLLIN) {
      /* read X events */
      XEvent event;
      wine_tsx11_lock();
      while (XPending(display)) {
	XNextEvent(display, &event);
	wine_tsx11_unlock();
	X11DRV_DD_ThreadEvent(display, &event);
	wine_tsx11_lock();
      }
      wine_tsx11_unlock();
    }
    if (fds[1].revents & POLLIN) {
      /* read command */
      int cmd;
      read(*input, &cmd, sizeof(cmd));
      if (X11DRV_DD_ThreadCmd(display, cmd)) break;
    }
  } while (TRUE);
  X11DRV_DD_DestroyWindow();
  return 0;
}

void X11DRV_DD_PostCmd(int cmd)
{
  write(X11DRV_DD_Pipe[1], &cmd, sizeof(cmd));
}

HRESULT X11DRV_DD_Sync(BOOL wait)
{
  if (X11DRV_DD_Paint || X11DRV_DD_Busy) {
    if (wait) {
      X11DRV_DD_PostCmd(DDMSG_SYNC);
      WaitForSingleObject(X11DRV_DD_SyncH, INFINITE);
    }
    else
      return DDERR_WASSTILLDRAWING;
  }
  return DD_OK;
}

void X11DRV_DD_Update(void)
{
  if (X11DRV_DD_Paint) return;
  if (!X11DRV_DD_PrimaryDIB) return;
  InterlockedIncrement(&X11DRV_DD_Paint);
  X11DRV_DD_PostCmd(DDMSG_UPDATE);
}

BOOL CALLBACK X11DRV_DD_Reparent(HWND hwnd, LPARAM lParam)
{
  WND *wnd = WIN_GetPtr(hwnd);
  if (!wnd) return TRUE;
  if (wnd == WND_OTHER_PROCESS) return TRUE; /* huh? */
  TRACE("hwnd=%08x, parent=%08x\n", hwnd, wnd->parent);
  if (!wnd->parent || wnd->parent == GetDesktopWindow()) {
    Display *display = thread_display();
    struct x11drv_win_data *data = wnd->pDriverData;
    Window win = data->whole_window;
    unsigned x, y;
    TRACE("reparenting %08lx\n", win);
    wine_tsx11_lock();
    if (root_window == DefaultRootWindow(gdi_display)) {
      /* turn on WS_EX_MANAGED? */
      /* fix window sizes? */
      x = data->whole_rect.left;
      y = data->whole_rect.top;
      XDefineCursor(display, win, X11DRV_GetCursor(display, X11DRV_NoCursor ? NULL : GlobalLock16(GetCursor())));
    }
    else {
      /* turn off WS_EX_MANAGED? */
      /* fix window sizes? */
      x = wnd->rectWindow.left;
      y = wnd->rectWindow.top;
      XUndefineCursor(display, win);
    }
    XReparentWindow(display, win, root_window,
                    wnd->rectWindow.left,
                    wnd->rectWindow.top);
    wine_tsx11_unlock();
  }
  WIN_ReleasePtr(wnd);
  return TRUE;
}

BOOL CALLBACK X11DRV_DD_Remap(HWND hwnd, LPARAM lParam)
{
  WND *wnd = WIN_GetPtr(hwnd);
  if (!wnd) return TRUE;
  if (wnd == WND_OTHER_PROCESS) return TRUE; /* huh? */
  TRACE("hwnd=%08x, parent=%08x\n", hwnd, wnd->parent);
  if (!wnd->parent || wnd->parent == GetDesktopWindow()) {
    if ((wnd->dwStyle & WS_VISIBLE) && !IsRectEmpty(&wnd->rectWindow)) {
      Display *display = thread_display();
      struct x11drv_win_data *data = wnd->pDriverData;
      Window win = data->whole_window;
      TRACE("%s %08lx\n", lParam ? "mapping" : "unmapping", win);
      wine_tsx11_lock();
#ifdef HAVE_OPENGL
      if (data->client_window == glXGetCurrentDrawable()) {
        TRACE("OpenGL context bound to window, ignoring\n");
      } else
#endif
      if (lParam)
        XMapWindow(display, win);
      else
        XUnmapWindow(display, win);
      wine_tsx11_unlock();
    }
  }
  WIN_ReleasePtr(wnd);
  return TRUE;
}

void X11DRV_DD_CreateDesktop(unsigned int width, unsigned int height)
{
  TRACE("(%d,%d)\n", width, height);
  if (!X11DRV_DD_RefCount++) {
    HWND desktop = GetDesktopWindow();
    RECT rect;

    if (X11DRV_DD_Thread) {
      ERR("thread already exist\n");
      return;
    }

    /* initialize new desktop thread */
    if (pipe(X11DRV_DD_Pipe) < 0) {
      ERR("pipe failed, error=%s\n", strerror(errno));
      return;
    }
    X11DRV_DD_SyncH = CreateEventA(NULL, FALSE, FALSE, NULL);
    X11DRV_DD_Thread = CreateThread(NULL, 0,
				    X11DRV_DD_ThreadProc,
				    X11DRV_DD_Pipe, 0, NULL);
    write(X11DRV_DD_Pipe[1], &width, sizeof(width));
    write(X11DRV_DD_Pipe[1], &height, sizeof(height));
    WaitForSingleObject(X11DRV_DD_SyncH, INFINITE);

    /* install new root window */
    X11DRV_DD_OldRootWindow = root_window;
    if (root_window == DefaultRootWindow(gdi_display)) {
      WND *deskwnd = WIN_GetPtr(desktop);
      struct x11drv_win_data *deskdata = deskwnd->pDriverData;
      root_window = X11DRV_DD_PrimaryClient;
      deskdata->whole_window = deskdata->client_window = root_window;
      WIN_ReleasePtr(deskwnd);
      /* FIXME: should probably reparent all top-level windows here,
       * but for now, only reparent windows owned by current thread */
      TSXSync(gdi_display, False);
      EnumThreadWindows(GetCurrentThread(), X11DRV_DD_Reparent, 0);
      TSXSync(thread_display(), False);
    }
    else {
      FIXME("can't resize desktop window yet\n");
      /* not necessary to change root */
    }
    /* change system metrics */
    screen_width = width;
    screen_height = height;
    /* change desktop window size */
    SetRect(&rect, 0, 0, screen_width, screen_height);
    WIN_SetRectangles(GetDesktopWindow(), &rect, &rect);
  }

  if (X11DRV_DD_DeviceWnd) {
    /* FIXME: should probably unmap all top-level windows here,
     * but for now, only unmap windows owned by current thread */
    TSXSync(gdi_display, False);
    EnumThreadWindows(GetCurrentThread(), X11DRV_DD_Remap, FALSE);
    TSXSync(thread_display(), False);
  }
}

void X11DRV_DD_DestroyDesktop(void)
{
  if (!X11DRV_DD_Thread) {
    ERR("thread doesn't exist\n");
    return;
  }
  TRACE("()\n");

  if (X11DRV_DD_DeviceWnd) {
    WND *win = WIN_GetPtr( X11DRV_DD_DeviceWnd );

    if (win && win != WND_OTHER_PROCESS) {
      /* restore old cooperative window position */
      if (X11DRV_DD_DeviceWndOK)
        WIN_SetRectangles(X11DRV_DD_DeviceWnd,
                          &X11DRV_DD_OldWindowRect,
                          &X11DRV_DD_OldClientRect);
      X11DRV_DD_DeviceWndOK = 0;

      WIN_ReleasePtr(win);
    }

    /* FIXME: should probably map all top-level windows here,
     * but for now, only unmap windows owned by current thread */
    TSXSync(gdi_display, False);
    EnumThreadWindows(GetCurrentThread(), X11DRV_DD_Remap, TRUE);
    TSXSync(thread_display(), False);
  }

  if (!--X11DRV_DD_RefCount) {
    HWND desktop = GetDesktopWindow();
    RECT rect;

    /* change system metrics */
    screen_width = default_screen_width;
    screen_height = default_screen_height;
    /* change desktop window size */
    SetRect(&rect, 0, 0, screen_width, screen_height);
    WIN_SetRectangles(desktop, &rect, &rect);

    /* restore old root window */
    root_window = X11DRV_DD_OldRootWindow;
    if (root_window == DefaultRootWindow(gdi_display)) {
      WND *deskwnd = WIN_GetPtr(desktop);
      struct x11drv_win_data *deskdata = deskwnd->pDriverData;
      deskdata->whole_window = deskdata->client_window = root_window;
      WIN_ReleasePtr(deskwnd);

      /* FIXME: should probably reparent all top-level windows here,
       * but for now, only reparent windows owned by current thread */
      TSXSync(gdi_display, False);
      EnumThreadWindows(GetCurrentThread(), X11DRV_DD_Reparent, 0);
      TSXSync(thread_display(), False);
    }
    else {
      FIXME("can't resize desktop window yet\n");
      /* not necessary to change root */
    }
    X11DRV_DD_OldRootWindow = 0;

    /* kill desktop thread */
    X11DRV_DD_PostCmd(DDMSG_QUIT);
    WaitForSingleObject(X11DRV_DD_Thread, INFINITE);
    CloseHandle(X11DRV_DD_Thread);
    X11DRV_DD_Thread = 0;
    CloseHandle(X11DRV_DD_SyncH);
    close(X11DRV_DD_Pipe[1]);
    close(X11DRV_DD_Pipe[0]);
  }
}
#endif

void X11DRV_DD_UnmapNotify(HWND hwnd, XUnmapEvent *event)
{
  if (!hwnd) return;
  if (hwnd == X11DRV_DD_DeviceWnd && X11DRV_DD_PrimaryGbl) {
    if (!X11DRV_DD_DeviceWndOK) {
      RECT rect;

      /* save old cooperative window position */
      WIN_GetRectangles(X11DRV_DD_DeviceWnd,
                        &X11DRV_DD_OldWindowRect,
                        &X11DRV_DD_OldClientRect);
      X11DRV_DD_DeviceWndOK = 1;

      /* make the cooperative window "fullscreen" */
      rect.left = 0; rect.right = X11DRV_DD_PrimaryGbl->wWidth;
      rect.top = 0; rect.bottom = X11DRV_DD_PrimaryGbl->wHeight;
      WIN_SetRectangles(X11DRV_DD_DeviceWnd, &rect, &rect);
    }
  }
}

static void SetPrimaryDIB(HBITMAP hBmp)
{
  X11DRV_DD_PrimaryDIB = hBmp;
  if (hBmp) {
    BITMAPOBJ *bmp;
    bmp = (BITMAPOBJ *)GDI_GetObjPtr( hBmp, BITMAP_MAGIC );
    if (!bmp->physBitmap) X11DRV_CreateBitmap(hBmp);
    X11DRV_DD_PrimaryDrawable = (Pixmap)bmp->physBitmap;
    GDI_ReleaseObj( hBmp );
  } else {
    X11DRV_DD_PrimaryDrawable = 0;
  }
}

void X11DRV_EnableGrabs(void)
{
  TRACE("\n");
  if (!dxgrab) return;
  if (X11DRV_DD_PrimaryFrame) {
    X11DRV_DD_PostCmd(DDMSG_GRAB);
    return;
  }
}

void X11DRV_DisableGrabs(void)
{
  TRACE("\n");
  if (!dxgrab) return;
  if (X11DRV_DD_PrimaryFrame) {
    X11DRV_DD_PostCmd(DDMSG_UNGRAB);
    return;
  }
}

void X11DRV_RetargetGrabs(void)
{
  TRACE("\n");
  if (!dxgrab) return;
  if (X11DRV_DD_PrimaryFrame) return;
  /* not needed anymore */
}

static DDHAL_DDCALLBACKS hal_ddcallbacks;

#ifdef DO_DIB
HRESULT X11DRV_DD_CreateDIB(LPDDRAWI_DDRAWSURFACE_LCL lcl, LPDDSURFACEDESC desc)
{
  LPDDRAWI_DDRAWSURFACE_GBL gbl = lcl->lpGbl;
  LPDDRAWI_DDRAWSURFACE_GBL_MORE gmore = GET_LPDDRAWSURFACE_GBL_MORE(gbl);
  BITMAPINFO* b_info;
  UINT usage;
  HDC ddc;
  LPVOID bitmap_data;
  DWORD pitch;

  if (gmore->hKernelSurface) return DD_OK; /* already allocated */

  switch (desc->ddpfPixelFormat.u1.dwRGBBitCount) {
  case 16:
  case 32:
    /* Allocate extra space to store the RGB bit masks. */
    b_info = (BITMAPINFO*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
				    sizeof(BITMAPINFOHEADER)
				    + 3 * sizeof(DWORD));
    break;

  case 24:
    b_info = (BITMAPINFO*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
				    sizeof(BITMAPINFOHEADER));
    break;

  default:
    /* Allocate extra space for a palette. */
    b_info = (BITMAPINFO*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
				    sizeof(BITMAPINFOHEADER)
				    + sizeof(RGBQUAD)
				    * (1 << desc->ddpfPixelFormat.u1.dwRGBBitCount));
    break;
  }

  b_info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  b_info->bmiHeader.biWidth = desc->dwWidth;
  b_info->bmiHeader.biHeight = -desc->dwHeight;
  b_info->bmiHeader.biPlanes = 1;
  b_info->bmiHeader.biBitCount = desc->ddpfPixelFormat.u1.dwRGBBitCount;

  if ((desc->ddpfPixelFormat.u1.dwRGBBitCount != 16)
      && (desc->ddpfPixelFormat.u1.dwRGBBitCount != 32))
    b_info->bmiHeader.biCompression = BI_RGB;
  else
    b_info->bmiHeader.biCompression = BI_BITFIELDS;

  b_info->bmiHeader.biSizeImage
    = (desc->ddpfPixelFormat.u1.dwRGBBitCount / 8)
    * desc->dwWidth * desc->dwHeight;

  b_info->bmiHeader.biXPelsPerMeter = 0;
  b_info->bmiHeader.biYPelsPerMeter = 0;
  b_info->bmiHeader.biClrUsed = 0;
  b_info->bmiHeader.biClrImportant = 0;

  switch (desc->ddpfPixelFormat.u1.dwRGBBitCount) {
  case 16:
  case 32:
    {
      DWORD *masks = (DWORD *) &(b_info->bmiColors);

      usage = 0;
      masks[0] = desc->ddpfPixelFormat.u2.dwRBitMask;
      masks[1] = desc->ddpfPixelFormat.u3.dwGBitMask;
      masks[2] = desc->ddpfPixelFormat.u4.dwBBitMask;
    }
    break;

  case 24:
    /* Nothing to do */
    usage = DIB_RGB_COLORS;
    break;

  default:
    /* Don't know palette */
    usage = 0;
    break;
  }

  ddc = CreateDCA("DISPLAY", NULL, NULL, NULL);
  if (ddc == 0) {
    HeapFree(GetProcessHeap(), 0, b_info);
    return HRESULT_FROM_WIN32(GetLastError());
  }

  pitch = (desc->dwFlags & DDSD_PITCH) ? desc->u1.lPitch : 0;
  gmore->hKernelSurface
    = DIB_CreateDIBSection(ddc, b_info, usage, &bitmap_data, 0,
			   (DWORD)desc->lpSurface, pitch);
  DeleteDC(ddc);
  if (!gmore->hKernelSurface) {
    ERR("CreateDIBSection failed!\n");
    HeapFree(GetProcessHeap(), 0, b_info);
    return HRESULT_FROM_WIN32(GetLastError());
  }

  TRACE("DIBSection at : %p\n", bitmap_data);

  gbl->fpVidMem = (FLATPTR)bitmap_data;
  gbl->u4.lPitch = pitch ? pitch : DIB_GetDIBWidthBytes(b_info->bmiHeader.biWidth,
							b_info->bmiHeader.biBitCount);

  HeapFree(GetProcessHeap(), 0, b_info);

  return S_OK;
}
#endif

static DWORD PASCAL X11DRV_DDHAL_DestroyDriver(LPDDHAL_DESTROYDRIVERDATA data)
{
  if (X11DRV_DD_Primary) {
    /* DestroySurface stuff */
#ifdef NEW_STUFF
    if (X11DRV_DD_DeviceWnd)
      X11DRV_DD_DestroyDesktop();
#endif
    X11DRV_DD_DeviceWnd = 0;
    X11DRV_DD_DeviceWindow = 0;
    X11DRV_DD_Primary = NULL;
    X11DRV_DD_PrimaryGbl = NULL;
#ifdef OLD_CRAP
    X11DRV_DD_PrimaryWnd = 0;
    X11DRV_DD_UserClass = 0;
#endif
    SetPrimaryDIB(0);
  }

  if (data->lpDD->dwFlags & DDRAWI_MODECHANGED) {
    DDHAL_SETMODEDATA mdata;

    /* restore default display mode */
    mdata.lpDD = data->lpDD;
    mdata.dwModeIndex = data->lpDD->dwModeIndexOrig;
    mdata.ddRVal = 0;
    mdata.SetMode = hal_ddcallbacks.SetMode;
    mdata.inexcl = 0;
    mdata.useRefreshRate = FALSE;
    if (mdata.SetMode)
      mdata.SetMode(&mdata);
  }
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_DDHAL_CreateSurface(LPDDHAL_CREATESURFACEDATA data)
{
  BOOL dib = TRUE, force_dib = FALSE;
  DWORD ret = DDHAL_DRIVER_NOTHANDLED;

  TRACE("caps=%lx\n", data->lpDDSurfaceDesc->ddsCaps.dwCaps);

  if (data->lpDDSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_EXECUTEBUFFER)
    dib = FALSE;
  else if (data->lpDDSurfaceDesc->ddpfPixelFormat.dwFlags & DDPF_FOURCC)
    dib = FALSE;

  if (data->lpDDSurfaceDesc->dwFlags & DDSD_LPSURFACE)
    /* the HEL shouldn't have to relocate client memory if it needs a DIB */
    force_dib = TRUE;
  else if (data->lpDDSurfaceDesc->ddsCaps.dwCaps & (DDSCAPS_PRIMARYSURFACE | DDSCAPS_BACKBUFFER))
    /* our update thread needs a DIB section */
    force_dib = TRUE;
  else if (data->lpDDSurfaceDesc->ddsCaps.dwCaps & (DDSCAPS_3DDEVICE | DDSCAPS_ZBUFFER))
    /* 3D stuff needs a DIB section */
    force_dib = TRUE;

#ifdef DO_DIB
  if (dib && force_dib) {
    DWORD c;
    for (c=0; c<data->dwSCnt; c++)
      X11DRV_DD_CreateDIB(data->lplpSList[c], data->lpDDSurfaceDesc);
  }
#endif

  if (data->lpDDSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_PRIMARYSURFACE) {
    if (X11DRV_DD_Primary) {
      ERR("created multiple primary surfaces\n");
      X11DRV_DisableGrabs();
#ifdef NEW_STUFF
      if (X11DRV_DD_DeviceWnd)
	X11DRV_DD_DestroyDesktop();
#endif
    }
    X11DRV_DD_Primary = *data->lplpSList;
    X11DRV_DD_PrimaryGbl = X11DRV_DD_Primary->lpGbl;
    /* get the cooperative window */
    if (data->lpDD->lpExclusiveOwner)
      X11DRV_DD_DeviceWnd = data->lpDD->lpExclusiveOwner->hWnd;
    else
      X11DRV_DD_DeviceWnd = 0;
    if (X11DRV_DD_DeviceWnd)
      X11DRV_DD_DeviceWindow = X11DRV_get_client_window(X11DRV_DD_DeviceWnd);
    else
      X11DRV_DD_DeviceWindow = 0;
#ifdef OLD_CRAP
    X11DRV_DD_PrimaryWnd = (HWND)X11DRV_DD_Primary->lpSurfMore->lpDDRAWReserved;
    X11DRV_DD_UserClass = GlobalFindAtomA("WINE_DDRAW");
#endif
    SetPrimaryDIB(GET_LPDDRAWSURFACE_GBL_MORE(X11DRV_DD_PrimaryGbl)->hKernelSurface);
#ifdef NEW_STUFF
    if (X11DRV_DD_DeviceWnd)
      X11DRV_DD_CreateDesktop(data->lpDDSurfaceDesc->dwWidth,
			      data->lpDDSurfaceDesc->dwHeight);
#endif

    X11DRV_EnableGrabs();
  }

  data->ddRVal = DD_OK;
  return ret;
}

static DWORD PASCAL X11DRV_DDHAL_CreatePalette(LPDDHAL_CREATEPALETTEDATA data)
{
  Display *display = gdi_display;
  TRACE("\n");
  if (visual->class == PseudoColor) {
    data->lpDDPalette->u1.dwReserved1 = TSXCreateColormap(display, root_window,
                                                          visual, AllocAll);
  }
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DDHAL_DDCALLBACKS hal_ddcallbacks = {
  sizeof(DDHAL_DDCALLBACKS),
  0x3ff, /* all callbacks are 32-bit */
  X11DRV_DDHAL_DestroyDriver,
  X11DRV_DDHAL_CreateSurface,
  NULL, /* SetColorKey */
  NULL, /* SetMode */
  NULL, /* WaitForVerticalBlank */
  NULL, /* CanCreateSurface */
  X11DRV_DDHAL_CreatePalette,
  NULL, /* GetScanLine */
  NULL, /* SetExclusiveMode */
  NULL  /* FlipToGDISurface */
};

static DWORD PASCAL X11DRV_DDHAL_DestroySurface(LPDDHAL_DESTROYSURFACEDATA data)
{
  LPDDRAWI_DDRAWSURFACE_GBL gbl = data->lpDDSurface->lpGbl;
  LPDDRAWI_DDRAWSURFACE_GBL_MORE gmore = GET_LPDDRAWSURFACE_GBL_MORE(gbl);

  TRACE("surface=%p\n", data->lpDDSurface);
  if (data->lpDDSurface == X11DRV_DD_Primary) {
    X11DRV_DisableGrabs();
#ifdef NEW_STUFF
    if (X11DRV_DD_DeviceWnd)
      X11DRV_DD_DestroyDesktop();
#endif
    X11DRV_DD_DeviceWnd = 0;
    X11DRV_DD_DeviceWindow = 0;
    X11DRV_DD_Primary = NULL;
    X11DRV_DD_PrimaryGbl = NULL;
#ifdef OLD_CRAP
    X11DRV_DD_PrimaryWnd = 0;
    X11DRV_DD_UserClass = 0;
#endif
    SetPrimaryDIB(0);
  }
#ifdef DO_DIB
  if (gmore->hKernelSurface) {
    /* DeleteObject(gmore->hKernelSurface); */
    /* gmore->hKernelSurface = 0; */
  }
#endif
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_DDHAL_Flip(LPDDHAL_FLIPDATA data)
{
  TRACE("surface=%p\n", data->lpSurfCurr);
  if (data->lpSurfCurr == X11DRV_DD_Primary) {
    /* data already flipped by ddraw, just need to update display */
#ifdef NEW_STUFF
    data->ddRVal = X11DRV_DD_Sync(data->dwFlags & DDFLIP_WAIT);
    if (data->ddRVal == DD_OK)
      X11DRV_DD_Update();
#endif
    return DDHAL_DRIVER_HANDLED;
  }
  return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_DDHAL_Lock(LPDDHAL_LOCKDATA data)
{
  TRACE("surface=%p\n", data->lpDDSurface);
  if (data->lpDDSurface == X11DRV_DD_Primary) {
#ifdef NEW_STUFF
#if 1
    data->ddRVal = DD_OK;
#else
    data->ddRVal = X11DRV_DD_Sync(data->dwFlags & DDLOCK_WAIT);
#endif
    if (data->ddRVal == DD_OK)
      InterlockedIncrement(&X11DRV_DD_Lock);
#endif
    return DDHAL_DRIVER_HANDLED;
  }
  return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_DDHAL_Unlock(LPDDHAL_UNLOCKDATA data)
{
  TRACE("surface=%p\n", data->lpDDSurface);
  if (data->lpDDSurface == X11DRV_DD_Primary) {
#ifdef NEW_STUFF
    InterlockedDecrement(&X11DRV_DD_Lock);
    X11DRV_DD_Update();
#endif
    data->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
  }
  return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_DDHAL_SetPalette(LPDDHAL_SETPALETTEDATA data)
{
  Colormap colormap = None;

  TRACE("surface=%p palette=%p\n", data->lpDDSurface, data->lpDDPalette);
  if (data->lpDDPalette)
    colormap = data->lpDDPalette->u1.dwReserved1;
  if (!colormap)
    colormap = X11DRV_PALETTE_PaletteXColormap;
  if (X11DRV_DD_Primary && data->lpDDSurface == X11DRV_DD_Primary) {
#ifdef NEW_STUFF
    if (visual->class == PseudoColor) {
      wine_tsx11_lock();
      XSetWindowColormap(gdi_display, X11DRV_DD_PrimaryClient, colormap);
      XSetWindowColormap(gdi_display, X11DRV_DD_PrimaryFrame, colormap);
      XFlush(gdi_display); /* update display immediately */
      wine_tsx11_unlock();
    } else
      X11DRV_DD_Update();
#endif
  }
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DDHAL_DDSURFACECALLBACKS hal_ddsurfcallbacks = {
  sizeof(DDHAL_DDSURFACECALLBACKS),
  0x3fff, /* all callbacks are 32-bit */
  X11DRV_DDHAL_DestroySurface,
  X11DRV_DDHAL_Flip,
  NULL, /* SetClipList */
  X11DRV_DDHAL_Lock,
  X11DRV_DDHAL_Unlock,
  NULL, /* Blt */
  NULL, /* SetColorKey */
  NULL, /* AddAttachedSurface */
  NULL, /* GetBltStatus */
  NULL, /* GetFlipStatus */
  NULL, /* UpdateOverlay */
  NULL, /* SetOverlayPosition */
  NULL, /* reserved4 */
  X11DRV_DDHAL_SetPalette
};

static DWORD PASCAL X11DRV_DDHAL_DestroyPalette(LPDDHAL_DESTROYPALETTEDATA data)
{
  Colormap pal = data->lpDDPalette->u1.dwReserved1;
  TRACE("palette=%p\n", data->lpDDPalette);
  if (pal) TSXFreeColormap(gdi_display, pal);
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_DDHAL_SetPaletteEntries(LPDDHAL_SETENTRIESDATA data)
{
  TRACE("palette=%p\n", data->lpDDPalette);
  X11DRV_DDHAL_SetPalEntries(data->lpDDPalette->u1.dwReserved1,
			     data->dwBase, data->dwNumEntries,
			     data->lpEntries);
  if (data->lpDDPalette->dwFlags & DDPCAPS_PRIMARYSURFACE) {
#ifdef NEW_STUFF
    if (!data->lpDDPalette->u1.dwReserved1)
      X11DRV_DD_Update();
#endif
  }
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DDHAL_DDPALETTECALLBACKS hal_ddpalcallbacks = {
  sizeof(DDHAL_DDPALETTECALLBACKS),
  0x3, /* all callbacks are 32-bit */
  X11DRV_DDHAL_DestroyPalette,
  X11DRV_DDHAL_SetPaletteEntries
};

static X11DEVICE x11device = {
  NULL
};

static DWORD PASCAL X11DRV_DDHAL_GetDriverInfo(LPDDHAL_GETDRIVERINFODATA data)
{
  LPX11DRIVERINFO info = x11device.lpInfo;
  while (info) {
    if (IsEqualGUID(&data->guidInfo, info->lpGuid)) {
      DWORD dwSize = info->dwSize;
      data->dwActualSize = dwSize;
      if (data->dwExpectedSize < dwSize) dwSize = data->dwExpectedSize;
      memcpy(data->lpvData, info->lpvData, dwSize);
      data->ddRVal = DD_OK;
      return DDHAL_DRIVER_HANDLED;
    }
    info = info->lpNext;
  }
  data->ddRVal = DDERR_CURRENTLYNOTAVAIL;
  return DDHAL_DRIVER_HANDLED;
}

static DDHALINFO hal_info = {
  sizeof(DDHALINFO),
  &hal_ddcallbacks,
  &hal_ddsurfcallbacks,
  &hal_ddpalcallbacks,
  {	/* vmiData */
   0	 /* fpPrimary */
  },
  {	/* ddCaps (only stuff the HAL implements here) */
   sizeof(DDCORECAPS),							/* dwSize */
   DDCAPS_GDI | DDCAPS_PALETTE,						/* dwCaps */
   DDCAPS2_CERTIFIED | DDCAPS2_NONLOCALVIDMEM | DDCAPS2_NOPAGELOCKREQUIRED |
   DDCAPS2_WIDESURFACES | DDCAPS2_FLIPNOVSYNC,				/* dwCaps2 */
   0,									/* dwCKeyCaps */
   0,									/* dwFXCaps */
   0,									/* dwFXAlphaCaps */
   DDPCAPS_8BIT | DDPCAPS_PRIMARYSURFACE,				/* dwPalCaps */
   0,									/* dwSVCaps */
   0,									/* dwAlphaBltConstBitDepths */
   0,									/* dwAlphaBltPixelBitDepths */
   0,									/* dwAlphaBltSurfaceBitDepths */
   0,									/* dwAlphaOverlayBltConstBitDepths */
   0,									/* dwAlphaOverlayBltPixelBitDepths */
   0,									/* dwAlphaOverlayBltSurfaceBitDepths */
   0,									/* dwZBufferBitDepths */
   16*1024*1024,							/* dwVidMemTotal */
   16*1024*1024,							/* dwVidMemFree */
   0,									/* dwMaxVisibleOverlays */
   0,									/* dwCurrVisibleOverlays */
   0,									/* dwNumFourCCCodes */
   0,									/* dwAlignBoundarySrc */
   0,									/* dwAlignSizeSrc */
   0,									/* dwAlignBoundaryDest */
   0,									/* dwAlignSizeDest */
   0,									/* dwAlignStrideAlign */
   {0},									/* dwRops */
   {									/* ddsCaps */
    DDSCAPS_BACKBUFFER | DDSCAPS_FLIP | DDSCAPS_FRONTBUFFER |
    DDSCAPS_OFFSCREENPLAIN | DDSCAPS_PALETTE | DDSCAPS_PRIMARYSURFACE |
    DDSCAPS_SYSTEMMEMORY | DDSCAPS_TEXTURE | DDSCAPS_VIDEOMEMORY |
    DDSCAPS_VISIBLE | DDSCAPS_LOCALVIDMEM | DDSCAPS_NONLOCALVIDMEM	/* dwCaps */
   },
   0,									/* dwMinOverlayStretch */
   0,									/* dwMaxOverlayStretch */
   0,									/* dwMinLiveVideoStretch */
   0,									/* dwMaxLiveVideoStretch */
   0,									/* dwMinHwCodecStretch */
   0,									/* dwMaxHwCodecStretch */
   0,									/* dwReserved1 */
   0,									/* dwReserved2 */
   0,									/* dwReserved2 */
   0,									/* dwSVBCaps */
   0,									/* dwSVBCKeyCaps */
   0,									/* dwSVBFXCaps */
   {0},									/* dwSVBRops */
   0,									/* dwVSBCaps */
   0,									/* dwVSBCKeyCaps */
   0,									/* dwVSBFXCaps */
   {0},									/* dwVSBRops */
   0,									/* dwSSBCaps */
   0,									/* dwSSBCKeyCaps */
   0,									/* dwSSBFXCaps */
   {0},									/* dwSSBRops */
   0,									/* dwMaxVideoPorts */
   0,									/* dwCurrVideoPorts */
   0									/* dwSVBCaps */
  },
  0,	/* dwMonitorFrequency */
  X11DRV_DDHAL_GetDriverInfo,
  0,	/* dwModeIndex */
  NULL,	/* lpdwFourCC */
  0,	/* dwNumModes */
  NULL,	/* lpModeInfo */
  DDHALINFO_ISPRIMARYDISPLAY | DDHALINFO_MODEXILLEGAL | DDHALINFO_GETDRIVERINFOSET, /* dwFlags */
  &x11device,
  0,	/* hInstance */
  0,	/* lpD3DGlobalDriverData */
  0,	/* lpD3DHALCallbacks */
  NULL	/* lpDDExeBufCallbacks */
};

static LPDDHALDDRAWFNS ddraw_fns;
static DWORD ddraw_ver;
static BOOL driver_created;

static void X11DRV_DDHAL_SetInfo(void)
{
  (ddraw_fns->lpSetInfo)(&hal_info, FALSE);
}

extern int __wine_hook_user_signal( void (*func)(CONTEXT*) );

static void X11DRV_UserSignal(CONTEXT*ctx)
{
  /* we could generalize this later */
#ifdef HAVE_OPENGL
#ifndef HAVE_OSMESA
  X11DRV_GLX_UserSignal(ctx);
#endif
#endif
}

void X11DRV_DD_PrimaryLock(void)
{
  DDHAL_LOCKDATA data;
  if (!X11DRV_DD_Primary) return;
  memset(&data, 0, sizeof(data));
  data.lpDDSurface = X11DRV_DD_Primary;
  hal_info.lpDDSurfaceCallbacks->Lock(&data);
}

void X11DRV_DD_PrimaryUnlock(void)
{
  DDHAL_UNLOCKDATA data;
  if (!X11DRV_DD_Primary) return;
  memset(&data, 0, sizeof(data));
  data.lpDDSurface = X11DRV_DD_Primary;
  hal_info.lpDDSurfaceCallbacks->Unlock(&data);
}

INT X11DRV_DCICommand(INT cbInput, const DCICMD *lpCmd, LPVOID lpOutData)
{
  TRACE("(%d,(%ld,%ld,%ld),%p)\n", cbInput, lpCmd->dwCommand,
	lpCmd->dwParam1, lpCmd->dwParam2, lpOutData);

  switch (lpCmd->dwCommand) {
  case DDNEWCALLBACKFNS:
    ddraw_fns = (LPDDHALDDRAWFNS)lpCmd->dwParam1;
    return TRUE;
  case DDVERSIONINFO:
    {
      LPDDVERSIONDATA lpVer = (LPDDVERSIONDATA)lpOutData;
      ddraw_ver = lpCmd->dwParam1;
      if (!lpVer) break;
      /* well, whatever... the DDK says so */
      lpVer->dwHALVersion = DD_RUNTIME_VERSION;
    }
    return TRUE;
  case DDGET32BITDRIVERNAME:
    {
      LPDD32BITDRIVERDATA lpData = (LPDD32BITDRIVERDATA)lpOutData;
      /* here, we could ask ddraw to load a separate DLL, that
       * would contain the 32-bit ddraw HAL */
      strcpy(lpData->szName,"x11drv");
      /* the entry point named here should initialize our hal_info
       * with 32-bit entry points (ignored for now) */
      strcpy(lpData->szEntryPoint,"DriverInit");
      lpData->dwContext = 0;
    }
    return TRUE;
  case DDCREATEDRIVEROBJECT:
    {
      LPDWORD lpInstance = (LPDWORD)lpOutData;

      /* FIXME: get x11drv's hInstance */

      if (!driver_created) {
        __wine_hook_user_signal(X11DRV_UserSignal);

#ifdef HAVE_LIBXXF86DGA2
	if (!X11DRV_XF86DGA2_CreateDriver(&hal_info))
#endif
	{
#ifdef HAVE_LIBXXF86VM
	  X11DRV_XF86VM_CreateDriver(&hal_info);
#endif
	}
#ifdef HAVE_OPENGL
#ifndef HAVE_OSMESA
	X11DRV_GLX_CreateDriver(&hal_info);
#else
	X11DRV_OSMesa_CreateDriver(&hal_info);
#endif
#endif
        driver_created = TRUE;
      }

      (ddraw_fns->lpSetInfo)(&hal_info, FALSE);
      *lpInstance = hal_info.hInstance;
    }
    return TRUE;
  }
  return 0;
}

void X11DRV_DDHAL_SwitchMode(DWORD dwModeIndex, LPVOID fb_addr, LPVIDMEM fb_mem)
{
  LPDDHALMODEINFO info = &hal_info.lpModeInfo[dwModeIndex];

  hal_info.dwModeIndex        = dwModeIndex;
  hal_info.dwMonitorFrequency = info->wRefreshRate;
  hal_info.vmiData.fpPrimary  = (FLATPTR)fb_addr;
  hal_info.vmiData.dwDisplayWidth  = info->dwWidth;
  hal_info.vmiData.dwDisplayHeight = info->dwHeight;
  hal_info.vmiData.lDisplayPitch   = info->lPitch;
  hal_info.vmiData.ddpfDisplay.dwSize = info->dwBPP ? sizeof(hal_info.vmiData.ddpfDisplay) : 0;
  hal_info.vmiData.ddpfDisplay.dwFlags = (info->wFlags & DDMODEINFO_PALETTIZED) ? DDPF_PALETTEINDEXED8 : 0;
  hal_info.vmiData.ddpfDisplay.u1.dwRGBBitCount = (info->dwBPP > 24) ? 24 : info->dwBPP;
  hal_info.vmiData.ddpfDisplay.u2.dwRBitMask = info->dwRBitMask;
  hal_info.vmiData.ddpfDisplay.u3.dwGBitMask = info->dwGBitMask;
  hal_info.vmiData.ddpfDisplay.u4.dwBBitMask = info->dwBBitMask;
  hal_info.vmiData.dwNumHeaps = fb_mem ? 1 : 0;
  hal_info.vmiData.pvmList = fb_mem;

  X11DRV_DDHAL_SetInfo();
}

void X11DRV_DDHAL_SetPalEntries(Colormap pal, DWORD dwBase, DWORD dwNumEntries,
				LPPALETTEENTRY lpEntries)
{
  XColor c;
  int n;

  if (pal) {
    wine_tsx11_lock();
    c.flags = DoRed|DoGreen|DoBlue;
    c.pixel = dwBase;
    for (n=0; n<dwNumEntries; n++,c.pixel++) {
      c.red   = lpEntries[n].peRed   << 8;
      c.green = lpEntries[n].peGreen << 8;
      c.blue  = lpEntries[n].peBlue  << 8;
      XStoreColor(gdi_display, pal, &c);
    }
    XFlush(gdi_display); /* update display immediately */
    wine_tsx11_unlock();
  }
}
