/* Wrapper for OpenGL includes...
     Copyright 1998 - Lionel Ulmer
     Copyright 2001 TransGaming Technologies, Inc.

   This wrapper is needed because Mesa uses also the CALLBACK / WINAPI
   constants.  It's also needed to work around certain common problems
   in Linux GL headers where key function pointers are not defined. */


#ifndef __WINE_WINE_GL_H
#define __WINE_WINE_GL_H

#ifndef __WINE_CONFIG_H
# error You must include config.h to use this header
#endif

#if defined(HAVE_OPENGL)

#include "ts_xlib.h"

/* As GLX relies on X, this is needed */
#define ENTER_GL() wine_tsx11_lock()
#define LEAVE_GL() wine_tsx11_unlock()

#undef APIENTRY
#undef CALLBACK
#undef WINAPI

#define XMD_H /* This is to prevent the Xmd.h inclusion bug to happen :-/ */
#include <GL/gl.h>
#include <GL/glu.h>
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>

/* Work around a gl.h vs glext.h snafu - gl.h may activate defines
   that prevent glext.h from giving us function pointer prototypes
   for the extensions below (which we need). If these defines are
   active, we have to define these prototypes ourselves. Of course
   this is further complicated by the fact that Mesa's gl.h header
   (if it's around instead of SGI's gl.h) includes glext.h itself.
   Hence the checks for GLEXT_VERSION for everything but
   GL_ARB_multitexture (prior to OpenGL 1.3, Mesa gl.h defined
   GL_ARB_multitexture itself, rather than leaving it up to glext.h
   - aak!) */
/* The nVidia gl.h also define extensions without giving us
   function pointer prototypes, and does declare OpenGL 1.3
   compatibility, but unlike Mesa, it doesn't include glext.h,
   so the check for GL_ARB_multitexture must check for this as well,
   sigh... */

#if defined(GL_ARB_multitexture) && (!defined(GL_GLEXT_VERSION) || !defined(GL_VERSION_1_3))
#define WINE_DDRAW_need_multitexture_pfns
#endif

#if defined(GL_EXT_secondary_color) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_secondary_color_pfns
#endif

#if defined(GL_ARB_texture_compression) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_texture_compression_pfns
#endif

#if defined(GL_EXT_vertex_weighting) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_vertex_weighting_pfns
#endif

#if defined(GL_NV_vertex_array_range) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_NV_VAR_pfns
#endif

#if defined(GL_NV_fence) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_NV_fence_pfns
#endif

#if defined(GL_NV_vertex_program) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_NV_vertex_program_pfns
#endif

#ifdef HAVE_GL_GLEXT_H
#include <GL/glext.h>
#endif

#ifndef APIENTRY
#define APIENTRY
#endif

#ifdef WINE_DDRAW_need_multitexture_pfns
typedef void (APIENTRY *PFNGLACTIVETEXTUREARBPROC)(GLenum texture);
typedef void (APIENTRY *PFNGLCLIENTACTIVETEXTUREARBPROC)(GLenum texture);
typedef void (APIENTRY *PFNGLMULTITEXCOORD2FARBPROC)(GLenum target, GLfloat u, GLfloat v);
#endif

#ifdef WINE_DDRAW_need_secondary_color_pfns
typedef void (APIENTRY * PFNGLSECONDARYCOLOR3UBEXTPROC) (GLubyte red, GLubyte green, GLubyte blue);
typedef void (APIENTRY * PFNGLSECONDARYCOLORPOINTEREXTPROC) (GLint size, GLenum type,
              GLsizei stride, GLvoid *pointer);
#endif

#ifdef WINE_DDRAW_need_texture_compression_pfns
typedef void (APIENTRY * PFNGLCOMPRESSEDTEXIMAGE2DARBPROC) (GLenum target, GLint level,
              GLenum internalformat, GLsizei width, GLsizei height, GLint border,
              GLsizei imageSize, const GLvoid *data);
typedef void (APIENTRY * PFNGLCOMPRESSEDTEXSUBIMAGE2DARBPROC) (GLenum target, GLint level,
              GLint xoffset, GLint yoffset, GLsizei width, GLsizei height,
              GLenum format, GLsizei imageSize, const GLvoid *data);
#endif


#ifdef WINE_DDRAW_need_vertex_weighting_pfns
typedef void (APIENTRY * PFNGLVERTEXWEIGHTFEXTPROC) (GLfloat weight);
typedef void (APIENTRY * PFNGLVERTEXWEIGHTFVEXTPROC) (const GLfloat *weight);
typedef void (APIENTRY * PFNGLVERTEXWEIGHTPOINTEREXTPROC) (GLsizei size, GLenum type, GLsizei stride, const GLvoid *pointer);
#endif


#ifdef WINE_DDRAW_need_NV_VAR_pfns
typedef void (APIENTRY * PFNGLFLUSHVERTEXARRAYRANGENVPROC) (void);
typedef void (APIENTRY * PFNGLVERTEXARRAYRANGENVPROC) (GLsizei size, const GLvoid *pointer);
#endif


#if defined(WINE_DDRAW_need_NV_vertex_program_pfns) || !defined(GL_NV_vertex_program)
typedef void (APIENTRY * PFNGLBINDPROGRAMNVPROC) (GLenum target, GLuint id);
typedef void (APIENTRY * PFNGLDELETEPROGRAMSNVPROC) (GLsizei n, const GLuint *programs);
typedef void (APIENTRY * PFNGLGENPROGRAMSNVPROC) (GLsizei n, GLuint *programs);
typedef void (APIENTRY * PFNGLLOADPROGRAMNVPROC) (GLenum target, GLuint id, GLsizei len, const GLubyte *program);
typedef void (APIENTRY * PFNGLPROGRAMPARAMETER4DNVPROC) (GLenum target, GLuint index, GLdouble x, GLdouble y, GLdouble z, GLdouble w);
typedef void (APIENTRY * PFNGLPROGRAMPARAMETER4DVNVPROC) (GLenum target, GLuint index, const GLdouble *v);
typedef void (APIENTRY * PFNGLPROGRAMPARAMETER4FNVPROC) (GLenum target, GLuint index, GLfloat x, GLfloat y, GLfloat z, GLfloat w);
typedef void (APIENTRY * PFNGLPROGRAMPARAMETER4FVNVPROC) (GLenum target, GLuint index, const GLfloat *v);
typedef void (APIENTRY * PFNGLPROGRAMPARAMETERS4DVNVPROC) (GLenum target, GLuint index, GLuint count, const GLdouble *v);
typedef void (APIENTRY * PFNGLPROGRAMPARAMETERS4FVNVPROC) (GLenum target, GLuint index, GLuint count, const GLfloat *v);
typedef void (APIENTRY * PFNGLVERTEXATTRIBPOINTERNVPROC) (GLuint index, GLint fsize, GLenum type, GLsizei stride, const GLvoid *pointer);
typedef void (APIENTRY * PFNGLGETPROGRAMPARAMETERFVNVPROC) (GLenum target, GLuint index, GLenum pname, GLfloat *params);
typedef void (APIENTRY * PFNGLGETPROGRAMPARAMETERDVNVPROC) (GLenum target, GLuint index, GLenum pname, GLdouble *params);
typedef void (APIENTRY * PFNGLGETVERTEXATTRIBPOINTERVNVPROC) (GLuint index, GLenum pname, GLvoid **pointer);
#endif


#if defined(GL_MAX_VERTEX_UNITS_ARB) && !defined(GL_ARB_vertex_blend)
typedef void (APIENTRY * PFNGLWEIGHTPOINTERARBPROC) (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
typedef void (APIENTRY * PFNGLVERTEXBLENDARBPROC) (GLint count);
#endif

#undef  XMD_H

#undef APIENTRY
#undef CALLBACK
#undef WINAPI

/* Redefines the constants */
#define CALLBACK    __stdcall
#define WINAPI      __stdcall
#define APIENTRY    WINAPI

#endif /* HAVE_OPENGL */

#endif /* __WINE_WINE_GL_H */
